(*  Title:      WellFormedness.thy
    ID:         $Id: WellFormedness.thy,v 1.3 2008/11/14 14:07:47 sowilo Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2007

    Note:       Proof of WF1 and WF2, well-formedness properties 
                for ASP redustions, i.e. no creation of future 
                (resp. activity) reference to non-existing future 
                (resp. activity) by the reduction rules
*)

header {* Well-formedness for ASP *}

theory WellFormedness imports Config begin

(* a term is welll formed if all references to futures or activities 
   map to an axisting future resp activity. A fut or act can be referenced either 
   from an active object or from a request. *)

constdefs
  wf_1a :: "configuration \<Rightarrow> bool"
  "wf_1a c == \<forall> A\<in> dom c. \<forall> E B. (c.\<^sub>AA = (E::general_context)\<up>(ActRef B)) \<longrightarrow> B\<in> dom c" 

  wf_1b  :: "configuration \<Rightarrow> bool"
  "wf_1b c == \<forall> A\<in> dom c. \<forall>fk\<in>dom (c.\<^sub>RA). \<forall> E B. (c.\<^sub>FA<fk>) =  (E:: general_context)\<up> (ActRef B) \<longrightarrow> B\<in> dom c"

  wf_2a :: "configuration \<Rightarrow> bool"
  "wf_2a c == \<forall> A\<in> dom c. \<forall> E f.(c.\<^sub>AA = (E::general_context)\<up>(FutRef f))
              \<longrightarrow>(\<exists> B\<in>dom c.  f\<in> dom(c.\<^sub>RB))" 

  wf_2b :: "configuration \<Rightarrow> bool"
  "wf_2b c == \<forall> A\<in> dom c. \<forall> fk\<in>dom (c.\<^sub>RA). \<forall> E f. (c.\<^sub>FA<fk>) = (E:: general_context)\<up>(FutRef f)
              \<longrightarrow>(\<exists> B\<in>dom c.  f\<in> dom(c.\<^sub>RB))"

  wf_conf :: "configuration \<Rightarrow> bool"
  "wf_conf c == wf_1a c \<and>wf_1b c \<and>wf_2a c \<and>wf_2b c"

subsection{*General Lemmas*}
(* reduction increases the set of defined activities *)
lemma actdom_and_red_inc: "c \<rightarrow>\<^sub>\<zeta>  c' \<Longrightarrow>dom c\<subseteq>dom c'";
apply (erule zeta.induct);
apply auto;
done;

(* reduction increases the set of defined requests *)
lemma futdom_and_red_inc: "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow>(\<forall> A\<in>dom c. dom(c.\<^sub>R A)\<subseteq>dom(c'.\<^sub>R A))";
apply (erule zeta.induct);
apply (auto simp: RequestQueue_def);
done;

lemma update_diff: "a\<noteq>b \<Longrightarrow> (m (a\<mapsto>f)) b = m b";    (* -->>I have not found it anywhere :-( *)
apply auto;
done;

lemma OneRequest_diff: "\<lbrakk>c A =Some(m,a);f\<noteq>f';f'\<in>dom m \<rbrakk> \<Longrightarrow> (c(A\<mapsto>(m (f\<mapsto>s),a))).\<^sub>FA<f'> =  (c.\<^sub>FA<f'>)"
apply (drule_tac f=s and m=m  in update_diff);
apply (subgoal_tac "c.\<^sub>FA<f'> = the (m f')");
apply (rule_tac m="m(f\<mapsto>s)" and a=a in OneRequestI);
apply simp+;
apply force;
apply (rule_tac m=m and a=a in OneRequestI);
apply simp;
apply force;
done;

lemma OneRequest_diff2: "\<lbrakk>A'\<in>dom c;f'\<in>dom (c.\<^sub>RA');A\<noteq>A' \<rbrakk> \<Longrightarrow> (c(A\<mapsto>t)).\<^sub>FA'<f'> =  (c.\<^sub>FA'<f'>)"
apply (frule_tac f=t and m=c  in update_diff);
apply (subgoal_tac "A'\<in> dom (c(A \<mapsto> t))")
apply (drule domD);
apply simp;
apply (erule exE)+;
apply (rule_tac m=a and a=b in OneRequestI);
apply simp+;
apply (drule domD);
apply (erule exE);
apply (subgoal_tac "(c .\<^sub>R A') = a");
apply (rule sym);
apply simp;
apply (rule_tac m=a and f=f' in OneRequestI);
apply (simp add: RequestQueue_def)+;
done;

lemma OneRequest_eq: " c A = Some (m,a)\<Longrightarrow> ((c(A\<mapsto>(m (f\<mapsto>s),a))).\<^sub>FA<f>)=s"
(*Hypothesis is not necessary but the theorem is always used this way *)
apply (rule_tac m="m(f\<mapsto>s)" and a=a in OneRequestI);
apply simp+;
done;

lemma ActiveObject_eq: " c A = Some (m,a)\<Longrightarrow> ((c(A\<mapsto>(m',a'))).\<^sub>AA)=a'"
(*Hypothesis is not necessary but the theorem is always used this way *)
by (force simp: ActiveObject_def)

lemma ActiveObject_diff: " A\<noteq>B \<Longrightarrow> ((c(B\<mapsto>t)).\<^sub>AA) = (c.\<^sub>AA)"
(*Hypothesis is not necessary but the theorem is always used this way *)
by (force simp: ActiveObject_def);


(* ----------------------    W F 1 ---------------------*)

subsection{* WF1 *}

(* a more precise lemma on the set of defined acts after reduction *)
lemma actdom_and_red_exact: "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow>dom c = dom c'\<or>(\<exists>!G. G\<in>(dom c'-dom c)\<and>(\<exists> a.(c' G) = Some(empty, a)))";
apply (erule zeta.induct);
apply (rule disjI1);
apply simp+;
apply (subgoal_tac "A:dom c");
defer;
apply force;
apply (rule disjI2);
apply (rule_tac a = G in ex1I);
apply (rule conjI);
apply force;
apply force;
apply (simp split add: split_if_asm)+;
apply force;
apply (rule disjI1);
defer;
apply (rule disjI1);
defer;
apply (rule disjI2);
apply (rule_tac a = G in ex1I);
apply (rule conjI);
apply auto;
apply (simp split add: split_if_asm)+;
apply force;
apply (simp split add: split_if_asm)+;
apply force;
apply (simp split add: split_if_asm)+;
done;

(* if an activity is referenced by an AO, was it the case before reduction? *)
lemma actref_and_red_ao: "\<lbrakk> c\<rightarrow>\<^sub>\<zeta> c';A \<in> dom c';(c'.\<^sub>A A)=(E::general_context)\<up>(ActRef B) \<rbrakk> 
  \<Longrightarrow>( (\<exists> A (E''::general_context) .A \<in> dom c \<and>c.\<^sub>A A = E''\<up>(ActRef B))\<or>
       (\<exists> G. G\<in> dom c\<and> (\<exists> f'\<in>dom (c.\<^sub>R G).
                    (\<exists>E'. (c.\<^sub>FG<f'>)= (E'::general_context)\<up>(Active(E\<up>(ActRef B)))\<or>
  (\<exists> (E'::general_context) C l E''::general_context. 
                    (c.\<^sub>FG<f'>)=E'\<up>(Upd (ActRef C) l (E''\<up>(ActRef B))))))))";
apply simp+;
apply (erule zeta.cases);
apply (simp split add: split_if_asm)+;
apply (unfold ActiveObject_def);
apply (simp split add: split_if_asm)+;
apply (rule disjI1);
apply force;
apply (rule disjI1);
apply force;

apply (simp split add: split_if_asm)+;
apply (rule disjI2);
apply (rule_tac x=Aa in exI);
apply (rule conjI,force);
apply (rule_tac x=fk in bexI);
apply (rule disjI1);
apply (rule_tac x=Ea in exI);
apply (unfold OneRequest_def);
apply force;
apply (force simp: RequestQueue_def)
apply (rule disjI1);
apply force;
apply (rule disjI1);
apply force;

(* request *)
apply (rule disjI1);
apply (rule_tac x=A in exI);
apply (simp split add: split_if_asm)+;
apply (rule conjI,force);
apply force;
apply force;
apply force;

(* self req *)
apply (rule disjI1);
apply (rule_tac x=A in exI);
apply (simp split add: split_if_asm)+;
apply (rule conjI,force);
apply force;
apply force;

(* upd ao *)
apply (simp split add: split_if_asm)+;
apply (case_tac E);
apply simp+;
apply (rule disjI1);apply force;
apply (rule disjI2);
apply (rule_tac x=A' in exI);
apply (rule conjI,force);
apply (rule_tac x=fk in bexI);
apply (rule disjI2);
apply (rule_tac x=Ea in exI);
apply (rule_tac x=Ba in exI);
apply (rule_tac x=li in exI);
apply (rule_tac x=general_context in exI);
apply force;
apply (force simp: RequestQueue_def)

apply simp+;
apply (rule disjI1);apply force;
apply (rule disjI1);apply force;
apply (rule disjI1);
apply (rule_tac x=A in exI);
apply (simp split add: split_if_asm)+;
apply auto;
done;

(* reference to activiities are not c reated by local red (with contexts) *)
lemma clocal_and_actref: "   a\<rightarrow>\<^bsub>c\<beta>\<^esub> a'\<longrightarrow>(\<forall> E G .(a'=((E::general_context)\<up>(ActRef G))\<longrightarrow>(\<exists> E'::general_context. a = E'\<up>(ActRef G))))"; 
apply (rule impI);
apply (erule cbeta.induct);
apply ((rule allI)+,rule impI);
apply (drule subst_context_Act_1);
apply (erule disjE);
apply (erule exE);
apply (rule_tac x = "(cCallL (cObj (Abs_FmapLabel(f,l)) T E')) l b" in exI);
apply simp+;
apply (subgoal_tac "f l =Some (E' \<up> (ActRef G))");
apply (drule_tac t = f and x= "(E' \<up> ActRef G)" in map_upd_triv);
apply simp+;
apply force;
apply (erule disjE,erule exE);
apply (rule_tac x = "cCallL E' l b" in exI);apply simp+;
apply (erule exE);
apply (rule_tac x = "cCallR (Obj f T) l E'" in exI);apply simp+;
apply ((rule allI)+,rule impI);
apply (subgoal_tac "(cObj (Abs_FmapLabel(f,l)) T cHole) \<up> a = E \<up> ActRef G")
apply (drule context_inc_Act);
apply (erule disjE, erule exE);
apply (rule_tac x = "cUpdR (Obj f T) l E1" in exI);
apply simp+;
apply (drule_tac x = "the (f l)" in spec);
apply (erule exE);
apply (rule_tac x = "cUpdL E2 l a" in exI);
apply simp+;
apply (subgoal_tac "f l = Some (the(f l))")
apply (drule_tac t = f  in map_upd_triv);
apply simp+;
apply force;
apply simp+;

apply ((rule allI)+,rule impI);
apply (case_tac "\<exists> E1::general_context. t=E1\<up>ActRef G");
apply (erule exE);
apply (drule_tac x=E1 in spec);
apply (drule_tac x=G in spec);
apply simp+;
apply (erule exE);
apply (rule_tac x = "E\<circ>E'" in exI);
apply simp+;
apply (drule context_inc_Act);
apply force;
done;

(* idem with the classical local red *)
lemma local_and_actref: "   a \<rightarrow>\<^sub>\<beta> (E::general_context)\<up>(ActRef G)\<Longrightarrow>
                                 \<exists> E'::general_context. a = E'\<up>(ActRef G)"; 
by (auto simp: clocal_and_actref cbeta_and_beta);

lemma diff_actrefs: "A\<noteq>B \<Longrightarrow> \<forall> E::general_context. ActRef A\<noteq>E \<up> (ActRef B)"
apply (rule allI);
apply (case_tac E);
apply auto;
done;
lemma context_actrefs: " ActRef A=E \<up> b\<Longrightarrow>E=cHole \<and>b=ActRef A"
apply (case_tac E);
apply auto;
done;

(* we define the set of activities created by a reduction, ie between 2 configurations *)
consts created_act:: "configuration \<Rightarrow> configuration \<Rightarrow> ActivityRef \<Rightarrow> bool"
defs
created_def:  "created_act c c' B \<equiv> c B = None \<and> (\<exists> a::dB . c' B = Some (empty,a))"

(* if an activity is referenced by a request, was it the case before reduction? *)
lemma actref_and_red_queue: "\<lbrakk> A\<in>dom c'; f\<in>dom(c'.\<^sub>RA); c'.\<^sub>FA<f> = (E::general_context)\<up>(ActRef B) ; c\<rightarrow>\<^sub>\<zeta> c'\<rbrakk>
  \<Longrightarrow>((\<exists> G\<in> dom c. \<exists> (E'::general_context). \<exists>f'\<in>dom (c.\<^sub>R G). (c.\<^sub>F G<f'> = E' \<up>(ActRef B)))\<or>
       (\<exists> G\<in>dom c. \<exists> (E'::general_context). c.\<^sub>A G = E'\<up>(ActRef B))\<or>
       ((created_act c c' B)  ))";
apply (erule zeta.cases);
apply (case_tac "A=Aa")
apply simp+;
apply (rule disjI1);
apply (case_tac "f=fk");

(*case f=fk *)
apply (simp add: OneRequest_def local_and_actref)+;
apply (rule_tac x = A in bexI);
apply (frule_tac a=a and E = E and G = B in local_and_actref);
apply (frule_tac f=f and s=a' in OneRequest_eq);apply simp;
apply (erule exE);
apply (rule_tac x = E' in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (unfold RequestQueue_def,force,fold RequestQueue_def);
apply force; 

(* case f\<noteq>fk *)
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (unfold RequestQueue_def,force,fold RequestQueue_def);
apply force; 
(* f\<noteq>fk done *)

apply (rule disjI1);
apply (frule OneRequestE);
apply force;apply assumption;
apply (erule exE)+;
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (unfold RequestQueue_def,force,fold RequestQueue_def);
apply force; 
(*local done*)


apply (case_tac "B = G");
apply (rule disjI2);
apply (rule disjI2);
apply simp;
apply (simp add: created_def);
apply (rule disjI1);
apply (case_tac "A=Aa");
apply (case_tac "f = fk");
apply simp+;

apply (rule_tac x = A in bexI);
prefer 2;apply force;
apply simp+;
apply (subgoal_tac "Ea \<up> ActRef G= E \<up> ActRef B");
apply (drule  contexts_and_dB);
apply (erule disjE)+;
apply (erule exE);
apply (rotate_tac -4);
apply (drule not_sym);
apply (simp add: diff_actrefs)+;
apply (erule disjE)+;
apply (erule exE);
apply (drule diff_actrefs)+;
apply (erule_tac x = E1 in allE)+;
apply simp;
apply (erule_tac x = "Active a" in allE)+;
apply (erule exE);
apply (rule_tac x = E2 in exI);
apply (rule_tac x = fk in bexI); prefer 2;
apply (unfold RequestQueue_def,force,fold RequestQueue_def);
apply (force simp: OneRequest_def);
apply (subgoal_tac "Aa\<in> dom (ca(Aa \<mapsto> (m(fk \<mapsto> Ea \<up> ActRef G), a'')))")
apply (drule_tac A=G and A'=Aa and t="(empty, a)" and f'=fk  in OneRequest_diff2)
apply (unfold RequestQueue_def,simp,fold RequestQueue_def);
apply force;
apply (force simp: OneRequest_def);
apply force;

(*easy cases for activity creation*)
(* 1 *)
apply (subgoal_tac "Aa\<noteq>G")
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply simp+;
apply (subgoal_tac "Aa\<in> dom (ca(Aa \<mapsto> (m(fk \<mapsto> Ea \<up> ActRef G), a'')))")
apply (frule_tac A=G and A'=Aa and t="(empty, a)" and f'=f  in OneRequest_diff2)
apply (unfold RequestQueue_def,simp,fold RequestQueue_def);
apply simp+;
apply (force simp: OneRequest_def);
apply (unfold RequestQueue_def,simp,fold RequestQueue_def);
apply (simp add: RequestQueue_def)+;
apply force;
apply force; 


(*2*)
apply (subgoal_tac "A\<noteq>G")
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (subgoal_tac "A\<in> dom (ca(Aa \<mapsto> (m(fk \<mapsto> Ea \<up> ActRef G), a'')))")
apply (frule_tac A=G and A'=A and t="(empty, a)" and f'=f  in OneRequest_diff2)
apply (unfold RequestQueue_def,simp,fold RequestQueue_def);
apply (force simp: OneRequest_def);
apply (subgoal_tac "A\<in> dom (ca(Aa \<mapsto> (m(fk \<mapsto> Ea \<up> ActRef G), a'')))")
apply (frule_tac A=G and A'=A and t="(empty, a)" and f'=f  in OneRequest_diff2)
apply simp+;
apply (force simp: OneRequest_def);
apply (unfold RequestQueue_def,simp,fold RequestQueue_def);
apply force;
apply (unfold RequestQueue_def,simp,fold RequestQueue_def);
apply force;
apply (subgoal_tac " (c' .\<^sub>R A)=(c .\<^sub>R A)")
apply (force simp: RequestQueue_def);
apply (simp split add: split_if_asm)+;
apply (force simp: RequestQueue_def);


(* REQUEST *)
apply (case_tac "A=Ba");
apply simp+;
apply (case_tac "f=fn");
apply simp+;
apply (subgoal_tac "(ca(A' \<mapsto> (m'(fk \<mapsto> Ea \<up> FutRef fn), a'))) Ba =Some (mb, a)")
prefer 2;apply force;
apply (drule_tac f=fn and s="Call a li p" and c="(ca(A' \<mapsto> (m'(fk \<mapsto> Ea \<up> FutRef fn), a')))" in OneRequest_eq)
apply (simp)+;
apply (rotate_tac 1,drule sym);
apply (frule context_call_1);
apply (erule disjE)+;
apply simp;
apply (rule disjI2);
apply (rule disjI1);
apply (erule exE);
apply (rule_tac x = Ba in bexI);
apply (rule_tac x = E' in exI);
apply (simp add: ActiveObject_def)
apply force;
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = A' in bexI);
apply (rule_tac x = "Ea \<circ> cCallR (ActRef Ba) li E'" in exI);
apply (rule_tac x = fk in bexI);
apply (simp add: OneRequest_def)+;
apply (force simp: RequestQueue_def);
apply force;

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;

apply (rule disjI1);
apply simp+;
apply (case_tac "A=A'");
apply (rule_tac x = A' in bexI);
apply (case_tac "fk=f");
apply simp+;
apply (subgoal_tac "A'\<in>dom(ca(A' \<mapsto> (m'(f \<mapsto> Ea \<up> FutRef fn), a')))",rotate_tac -1)
apply (drule_tac A=Ba and f'=f and t="(mb(fn \<mapsto> Call a li p), a)" in OneRequest_diff2);
apply (unfold RequestQueue_def,simp,fold RequestQueue_def);
apply force;
apply (frule_tac f=f and A=A' and s="Ea \<up> FutRef fn" in OneRequest_eq)
apply simp+;
apply (drule  contexts_and_dB);
apply (erule disjE, erule exE,case_tac E1, simp+)+;
apply (drule_tac x = "Call (ActRef Ba) li p" in spec, erule exE);
apply (rule_tac x = E2 in exI);
apply (rule_tac x = fk in bexI);
apply (force simp: ectxt2gctxt_and_fill OneRequest_def) ;
apply (force simp: RequestQueue_def);
apply force;
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (rule OneRequestI);
apply force;
apply simp+;
apply (rotate_tac 1,drule not_sym)
apply (drule OneRequestE)
apply simp+;
apply (force simp: RequestQueue_def);
apply force;

apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;


(* self request*)
apply (case_tac "A=A'");
apply simp+;
apply (case_tac "f=fn");
apply simp+;
apply (subgoal_tac "(ca(A' \<mapsto> (m'(fk \<mapsto> Ea \<up> FutRef fn), a))) A' =Some (m'(fk \<mapsto> Ea \<up> FutRef fn), a)")
prefer 2;apply force;
apply (drule_tac f=fn and s="Call a li p" and c="(ca(A' \<mapsto> (m'(fk \<mapsto> Ea \<up> FutRef fn), a)))" in OneRequest_eq)
apply (simp)+;
apply (frule context_call_1);
apply (erule disjE)+;
apply simp;
apply (rule disjI2);
apply (rule disjI1);
apply (rule_tac x = A' in bexI);
apply (erule exE)+;
apply (rule_tac x = E' in exI);
apply (unfold ActiveObject_def,simp+,fold ActiveObject_def);
apply force;

apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = A' in bexI);
apply (rule_tac x = "Ea \<circ> cCallR (ActRef A') li E'" in exI);
apply (rule_tac x = fk in bexI);
apply (unfold OneRequest_def,simp+,fold OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;


apply (case_tac "f=fk");
apply simp+;
apply (subgoal_tac "Ea \<up> FutRef fn = E \<up> ActRef B")
prefer 2;apply (force simp: OneRequest_def);
apply (drule  contexts_and_dB);
apply (erule disjE, erule exE,case_tac E1, simp+)+;
apply (drule_tac x = "Call (ActRef A') li p" in spec, erule exE);
apply (rule disjI1);
apply simp+;
apply (rule_tac x = A' in bexI);
apply (rule_tac x = E2 in exI);
apply (rule_tac x = fk in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (rule OneRequestI);
apply force;
apply (rotate_tac -6);
apply (frule_tac f=fk and f'=f in OneRequest_diff); 
apply force;
apply (unfold RequestQueue_def,force,fold RequestQueue_def);
apply (subgoal_tac "(ca(A' \<mapsto> (m'(fk \<mapsto> Ea \<up> FutRef fn), a))) A'=Some (m'(fk \<mapsto> Ea \<up> FutRef fn),a)",rotate_tac -1);
apply (frule_tac f=fn and f'=f in OneRequest_diff); 
apply force;
apply (force simp: RequestQueue_def)
apply (unfold OneRequest_def);apply simp+;
apply (force simp: RequestQueue_def);
apply force;
apply (force simp: RequestQueue_def);
apply (fold OneRequest_def)
apply force;

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;

(* AO update *)
apply (subgoal_tac "A\<noteq>G");
prefer 2;apply (force simp: RequestQueue_def);
apply (case_tac "A=A'");
apply (case_tac "B=G");
apply (rule disjI2);
apply (rule disjI2);
apply (simp add: created_def);
apply (rule disjI1);
apply (case_tac "f=fk");
prefer 2;
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply simp+;
apply (force simp: RequestQueue_def);
apply force;

apply (subgoal_tac "Ea \<up> (ActRef G) = E \<up> ActRef B");
prefer 2;
apply (subgoal_tac "(A' \<in> dom (ca(A' \<mapsto> (m'(fk \<mapsto> Ea \<up> ActRef G), a'))))",rotate_tac -1)
apply (drule_tac A=G in OneRequest_diff2);
apply (force simp: RequestQueue_def);
apply force;
apply (force simp: OneRequest_def);
apply force;
apply (drule  context_inc_Act);
apply (erule disjE);
apply (erule exE);
apply (rotate_tac 11,drule not_sym,rotate_tac -1);
apply (drule diff_actrefs);
apply (simp+);
apply (drule_tac x = " Upd (ActRef Ba) li b" in spec);
apply (erule exE);
apply (rule_tac x = A in bexI);
apply (rule_tac x = E2 in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;

(* future update *)
apply (rule disjI1);
apply safe;
apply (case_tac "A=Aa");
apply (case_tac "f=fk");
apply simp+;
apply (erule conjE);
apply (subgoal_tac " Ea \<up> ai =  E \<up> ActRef B")
apply (drule  contexts_and_dB);
apply (erule disjE, erule exE, simp+);
apply (rule_tac x = Ba in bexI);
apply (rule_tac x = E1 in exI);
apply (rule_tac x = fi in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;

apply (erule disjE, erule exE);
apply (rotate_tac -1,drule sym);
apply (drule context_actrefs); 
apply (erule conjE,simp+);
apply (rule_tac x = Ba in bexI);
apply (rule_tac x = "cHole" in exI);
apply (rule_tac x = fi in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;

apply (drule_tac x="FutRef fi" in spec,erule exE);
apply (rule_tac x = A in bexI);
apply (rule_tac x = E2 in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;
apply (force simp: OneRequest_def);
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;
apply (rule_tac x = A in bexI);
apply (rule_tac x = E in exI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def);
apply (force simp: RequestQueue_def);
apply force;
done;

(* reduction maintans WF1a+b *)
lemma subject_reduction_WF1[rule_format]: "c\<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> (wf_1a c\<and>wf_1b c  \<longrightarrow> wf_1a c'\<and>wf_1b c')";
apply (unfold wf_1a_def);
apply (unfold wf_1b_def);
apply (rule impI);
apply (unfold ActiveObject_def);
apply (erule conjE);
apply (rule conjI);
apply (rule ballI);
apply (rule allI)+;
apply (rule impI);
apply (frule_tac c=c in actref_and_red_ao);
apply (simp add: ActiveObject_def)
apply (unfold ActiveObject_def,force,fold ActiveObject_def);
apply (erule disjE)+;
apply (erule exE)+;
apply (drule_tac x=Aa in bspec);
apply simp;
apply simp;
apply (drule_tac x=E'' in spec);
apply (drule_tac x=B in spec);
apply (simp)+;
apply (drule actdom_and_red_inc);
apply blast;

apply ((erule exE),(erule conjE),(erule bexE),(erule exE));
apply (rotate_tac 2);
apply (erule disjE);
apply (drule_tac x=G in bspec);
apply force;
apply (drule_tac x=f' in bspec);
apply (simp add: RequestQueue_def)
apply (drule_tac x="E'\<circ>(cActive E)" in spec);
apply (drule_tac x=B in spec);
apply (simp)+;
apply (drule actdom_and_red_inc);
apply blast;
apply (erule exE)+;
apply (drule_tac x=G in bspec);
apply force;
apply (drule_tac x=f' in bspec);
apply force;
apply (drule_tac x="E'a\<circ>(cUpdR (ActRef C) l (E''))" in spec);
apply (drule_tac x=B in spec);
apply (simp)+;
apply (drule actdom_and_red_inc);
apply blast;

apply (rule ballI)+;
apply (rule allI)+;
apply (rule impI);
apply (frule_tac c=c and f=fk and E=E and B=B in actref_and_red_queue);
apply (assumption)+;
apply (erule disjE, (erule bexE)+,(erule exE)+, (erule bexE));
apply (rotate_tac 2);
apply (drule_tac x=G in bspec);  apply assumption;
apply (drule_tac x=f' in bspec); apply assumption;
apply (drule_tac x=E' in spec);
apply (drule_tac x=B in spec);
apply simp+;
apply (drule actdom_and_red_inc);
apply blast;
apply (erule disjE)
apply ((erule bexE),erule exE);
apply (drule_tac x=G in bspec);
apply assumption;
apply (drule_tac x=E' in spec);
apply (drule_tac x=B in spec);
apply simp+;
apply (drule actdom_and_red_inc);
apply (auto simp: created_def);
done;


(* ----------------   W F  2 --------------------  *)
subsection{* WF2 *}

(* if a future is refenrenced by an AO, was it the case bfore the reduction ? *)
lemma futref_and_red_ao: "\<lbrakk>c\<rightarrow>\<^sub>\<zeta> c';A\<in>dom c';c'.\<^sub>A A = (E::general_context)\<up>(FutRef f)\<rbrakk> 
  \<Longrightarrow>((\<exists> B\<in>dom c.\<exists> (E'::general_context) .c.\<^sub>A B = E'\<up>(FutRef f)) \<or>
      (\<exists> G\<in> dom c.\<exists> f'\<in>dom (c.\<^sub>RG).( \<exists> E' a. ((c.\<^sub>FG<f'>) = ((E'::general_context)\<up>Active(E\<up>(FutRef f)))) \<and>A\<notin>dom c)\<or>
         (\<exists> li D (E'::general_context) (E''::general_context). 
             ((c.\<^sub>FG<f'>)= E'\<up>(Upd (ActRef D) li (E''\<up>(FutRef f))) \<and>A\<notin>dom c ))))";
apply simp+;
apply (erule zeta.cases);
apply (rule disjI1);
apply (case_tac "A=Aa");
apply (force simp: ActiveObject_def)
apply (force simp: ActiveObject_def)

(* AO creation*)
apply (case_tac "A=Aa");
apply (rule disjI1);
apply (subgoal_tac "A\<noteq>G")
apply (force simp: ActiveObject_def)
apply force;
apply (case_tac "A=G");
apply (rule disjI2)
apply (rule_tac x=Aa in bexI)
apply (rule_tac x=fk in bexI)
apply (rule disjI1)
apply (rule conjI);
apply (rule_tac x=Ea in exI);
apply (force simp: OneRequest_def ActiveObject_def)
apply force;
apply (force simp: RequestQueue_def)
apply force;
apply (rule disjI1)
apply (force simp: ActiveObject_def)

(* REQUEST *)
apply (rule disjI1);
apply (case_tac "A=A'");
apply (force simp: ActiveObject_def)
apply (case_tac "A=B");
apply (force simp: ActiveObject_def)
apply (force simp: ActiveObject_def)

(* SELF REQUEST *)
apply (rule disjI1);
apply (case_tac "A=A'");
apply (force simp: ActiveObject_def)
apply (force simp: ActiveObject_def)

(* UPDATE AO *)
apply (case_tac "A=A'");
apply (rule disjI1);
apply (subgoal_tac "A\<noteq>G")
apply (force simp: ActiveObject_def)
apply force;
apply (case_tac "A=G");
apply (simp add: ActiveObject_def)+;
apply (drule sym,frule context_update);
apply (erule disjE)+;
apply simp+;
apply (erule exE);
apply (rule disjI1)
apply (force simp: ActiveObject_def)

apply (rule disjI2)
apply (rule_tac x=A' in bexI)
apply (rule_tac x=fk in bexI)
apply (rule disjI2)
apply (rule conjI);
apply (erule exE);
apply (rule_tac x=li in exI)
apply (rule_tac x=B in exI)
apply (rule_tac x=Ea in exI)
apply (rule_tac x=E' in exI)
apply (force simp: OneRequest_def)
apply blast;
apply (force simp: RequestQueue_def);
apply blast;

apply (rule disjI1);
apply (force simp: ActiveObject_def)

(* FUT UPDATE *)
apply (rule disjI1);
apply (case_tac "A=Aa") 
apply (force simp: ActiveObject_def)
apply (force simp: ActiveObject_def)
done;

(* local reduction does not create future reference, context reduction *)
lemma clocal_and_futref: "   a\<rightarrow>\<^bsub>c\<beta>\<^esub> a'\<longrightarrow>(\<forall> E fk .(a'=((E::general_context)\<up>(FutRef fk))\<longrightarrow>(\<exists> E'::general_context. a = E'\<up>(FutRef fk))))"; 
apply (rule impI);
apply (erule cbeta.induct);
apply ((rule allI)+,rule impI);
apply (drule subst_context_Fut_1);
apply (erule disjE);
apply (erule exE);
apply (rule_tac x = "cCallL (cObj (Abs_FmapLabel(f,l)) T E') l b" in exI);
apply simp+;
apply (subgoal_tac "f l =Some (E' \<up> (FutRef fk))");
apply (drule_tac t = f and x= "(E' \<up> FutRef fk)" in map_upd_triv);
apply simp+;
apply force;
apply (erule disjE);

apply (erule exE);
apply (rule_tac x = "cCallL E' l b" in exI);apply simp+;
apply (erule exE);
apply (rule_tac x = "cCallR (Obj f T) l E'" in exI);
apply simp+;

apply ((rule allI)+,rule impI);
apply (subgoal_tac "(cObj (Abs_FmapLabel(f,l)) T cHole) \<up> a = E \<up> FutRef fk")
apply (drule context_inc_Fut);
apply (erule disjE, erule exE);
apply (rule_tac x = "cUpdR (Obj f T) l E1" in exI);
apply simp+;
apply (drule_tac x = "the (f l)" in spec);
apply (erule exE);
apply (rule_tac x = "cUpdL E2 l a" in exI);
apply simp+;
apply (subgoal_tac "f l = Some (the(f l))")
apply (drule_tac t = f  in map_upd_triv);
apply simp+;
apply force;
apply simp+;

apply ((rule allI)+,rule impI);
apply (case_tac "\<exists> E1::general_context. t=E1\<up>FutRef fk");
apply (erule exE);
apply (drule_tac x=E1 in spec);
apply (drule_tac x=fk in spec);
apply simp+;
apply (erule exE);
apply (rule_tac x = "E\<circ>E'" in exI);
apply simp+;
apply (drule context_inc_Fut);
apply force;
done;

(* now classical reduction *)
lemma local_and_futref: "   \<lbrakk>a \<rightarrow>\<^sub>\<beta> a';a'=(E::general_context)\<up>(FutRef fk)\<rbrakk>\<Longrightarrow>
                                \<exists> E'::general_context. a = E'\<up>(FutRef fk)"; 
by (auto simp: clocal_and_futref cbeta_and_beta);

lemma diff_futrefs: "fk\<noteq>fl \<Longrightarrow> \<forall> E::general_context. FutRef fk\<noteq>E \<up> (FutRef fl)"
apply (rule allI);
apply (case_tac E);
apply auto;
done;
lemma diff_futactrefs: "\<forall> E::general_context. ActRef G\<noteq>E \<up> (FutRef fl)"
apply (rule allI);
apply (case_tac E);
apply auto;
done;
lemma diff_actfutrefs: "\<forall> E::general_context. FutRef fk\<noteq>E \<up> (ActRef G)"
apply (rule allI);
apply (case_tac E);
apply auto;
done;

(* we define the set of future created, inside a given activity by a reduction, 
  ie between 2 configurations *)
consts created_fut:: "configuration \<Rightarrow> configuration \<Rightarrow> ActivityRef \<Rightarrow> FutureRef \<Rightarrow> bool"
defs
fcreated_def:  "created_fut c c' A fk \<equiv> 
                          A \<in> dom c'
                          \<and> (\<forall> B\<in>dom c. fk\<notin> dom (c.\<^sub>RB)) 
                          \<and> fk\<in> dom (c'.\<^sub>RA)"

(* if a future is referenced by a request term after reduction, was it the case before ? *)
lemma futref_and_red_queue: "\<lbrakk> c\<rightarrow>\<^sub>\<zeta> c'; A\<in>dom c';f\<in>dom(c'.\<^sub>RA);(c'.\<^sub>FA<f>)=(E::general_context)\<up>(FutRef fk)\<rbrakk>
  \<Longrightarrow>((\<exists> G\<in>dom c. \<exists>f'\<in> dom (c.\<^sub>RG). \<exists> (E'::general_context). (c.\<^sub>FG <f'>) = E' \<up>(FutRef fk)) \<or>
       (\<exists> G\<in>dom c . \<exists> (E'::general_context). (c.\<^sub>A G) = E'\<up>(FutRef fk)) \<or>
       (\<exists> B li  b (E'::general_context). ((c.\<^sub>FA <f>)=E'\<up>(Call (ActRef B) li b))\<and>(created_fut c c' B fk)  ))";
apply (erule zeta.cases);

(* LOCAL *)
apply (case_tac "A=Aa");
apply (case_tac "f=fka");
prefer 2;
apply (rule disjI1);
apply (force simp: OneRequest_def RequestQueue_def)

apply (simp add: local_and_futref)+;
apply (drule_tac a=a and E = E and fk = fk in local_and_futref);
apply (force simp: OneRequest_def)
apply (rule disjI1);
apply (force simp: OneRequest_def RequestQueue_def)

apply (rule disjI1);
apply (force simp: OneRequest_def RequestQueue_def)


(* AO creation *)
apply (subgoal_tac "A\<noteq>G")
apply (rule disjI1);
apply (case_tac "A=Aa");
apply (case_tac "f=fka");
apply simp+;
apply (rule_tac x = A in bexI);
apply simp+;
apply (subgoal_tac "Ea \<up> ActRef G= E \<up> FutRef fk");
apply (drule  contexts_and_dB);
apply (erule disjE)+;
apply (erule exE);
apply (simp add: diff_futactrefs);
apply (erule disjE)+;
apply (erule exE);
apply (rotate_tac -1,drule sym);apply (simp add: diff_actfutrefs)+;
apply (erule_tac x = "Active a" in allE)+;
apply (erule exE);
apply (rule_tac x = f in bexI);
apply (rule_tac x = E2 in exI);
apply (force simp: OneRequest_def RequestQueue_def)
apply (force simp: OneRequest_def RequestQueue_def)

apply (force simp: OneRequest_def RequestQueue_def)
apply blast;

apply (force simp: OneRequest_def RequestQueue_def)
apply (force simp: OneRequest_def RequestQueue_def)
apply (force simp: OneRequest_def RequestQueue_def)


(* request *)
apply (case_tac "A=B");
apply (case_tac "f=fn");
apply (subgoal_tac "E \<up> FutRef fk = Call a li p")
apply (frule context_call_1);
apply (erule disjE)+;
apply simp;
(* futref in called AO *)
apply (rule disjI2);
apply (rule disjI1);
apply (erule exE)+;
apply simp;
apply (rule_tac x = B in bexI);
apply (rule_tac x = E' in exI);
apply (force simp: ActiveObject_def)
apply blast;

(* futref in param *)
apply (rule disjI1);
apply (erule exE)+;
apply (rule_tac x = A' in bexI);
apply (rule_tac x = fka in bexI);
apply (rule_tac x = "Ea\<circ>(cCallR (ActRef B) li E')" in exI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply blast;
apply (force simp: OneRequest_def RequestQueue_def)

(* ELSE *)
apply (rule disjI1);
apply (force simp: OneRequest_def RequestQueue_def)

apply (case_tac "A=A'");
apply (case_tac "f=fka");
apply (subgoal_tac "Ea \<up> FutRef fn = E \<up> FutRef fk")
apply (frule context_inc_Fut);
apply (erule disjE)+;
apply (erule exE);
apply (subgoal_tac "fn=fk")
apply simp;
apply (drule fixpoint_context);
apply (rule disjI2);
apply (rule disjI2);
apply (rule_tac x = B in exI);
apply (rule conjI);
apply (force simp: OneRequest_def RequestQueue_def)
apply (simp add: fcreated_def)+;
apply (rule conjI)
apply (force simp: RequestQueue_def)
apply (force simp: RequestQueue_def)
apply (force simp: diff_futrefs);

apply (drule_tac x="Call (ActRef B) li p" in spec)
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = A' in bexI);
apply (rule_tac x = f in bexI);
apply (rule_tac x = E2 in exI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply blast;

apply (force simp: OneRequest_def)

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply blast;

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply force;

(* SELF REQUEST *)
apply (case_tac "A=A'");
apply (case_tac "f=fn");
apply (subgoal_tac "E \<up> FutRef fk = Call a li p")
apply (frule context_call_1);
apply (erule disjE)+;
apply simp;
(* futref in called AO *)
apply (rule disjI2);
apply (rule disjI1);
apply (erule exE)+;
apply simp;
apply (rule_tac x = A' in bexI);
apply (rule_tac x = E' in exI);
apply (force simp: ActiveObject_def)
apply blast;

(* futref in param *)
apply (rule disjI1);
apply (erule exE)+;
apply (rule_tac x = A' in bexI);
apply (rule_tac x = fka in bexI);
apply (rule_tac x = "Ea\<circ>(cCallR (ActRef A') li E')" in exI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply blast;
apply (force simp: OneRequest_def RequestQueue_def)

(* ELSE *)
apply (case_tac "f=fka");
apply (subgoal_tac "Ea \<up> FutRef fn = E \<up> FutRef fk")
apply (frule context_inc_Fut);
apply (erule disjE)+;
apply (erule exE);
apply (subgoal_tac "fn=fk")
apply simp;
apply (drule fixpoint_context);
apply (rule disjI2);
apply (rule disjI2);
apply (rule_tac x = A' in exI);
apply (rule conjI);
apply (force simp: OneRequest_def RequestQueue_def)
apply (simp add: fcreated_def)+;
apply (rule conjI)
apply (force simp: RequestQueue_def)
apply (force simp: RequestQueue_def)
apply (force simp: diff_futrefs);

apply (drule_tac x="Call (ActRef A') li p" in spec)
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = A' in bexI);
apply (rule_tac x = f in bexI);
apply (rule_tac x = E2 in exI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply blast;

apply (force simp: OneRequest_def)

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply blast;

apply (rule disjI1);
apply (rule_tac x = A in bexI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def)
apply (force simp: RequestQueue_def)
apply force;

(* AO update*)
apply (case_tac "A=G");
prefer 2;
apply (rule disjI1);
apply (case_tac "A=A'");
apply (rule_tac x = A in bexI);

apply (case_tac "f=fka");
apply (subgoal_tac " Ea \<up> ActRef G = E \<up> FutRef fk");
prefer 2;apply (force simp: OneRequest_def) ;
apply (drule context_inc_Fut);
apply (erule disjE);
apply (erule exE,simp+);
apply (force simp: diff_futactrefs);
apply (drule_tac x="Upd (ActRef B) li b" in spec,erule exE);
apply (rule_tac x = f in bexI);apply (force simp: OneRequest_def) ;apply (force simp: RequestQueue_def);
apply (rule_tac x = f in bexI);apply (force simp: OneRequest_def) ;apply (force simp: RequestQueue_def);
apply force;
apply (rule_tac x = A in bexI);
apply (rule_tac x = f in bexI);apply (force simp: OneRequest_def) ;apply (force simp: RequestQueue_def);
apply force;

(* A=G impossible *)
apply (force simp: RequestQueue_def);

(* future uodate *) 
apply simp+;
apply (rule disjI1);
apply (case_tac "A=Aa");
apply (case_tac "f=fka");
apply simp+;
apply (subgoal_tac "Ea \<up> ai =  E \<up> FutRef fk");

apply (drule  context_inc_Fut);
apply (erule disjE, erule exE, simp+);
apply (rule_tac x = B in bexI);
apply (rule_tac x = fi in bexI);
apply (force simp: OneRequest_def) ;
apply (force simp: RequestQueue_def);
apply blast;

apply (drule_tac x="FutRef fi" in spec); 
apply (erule exE);
apply (rule_tac x = A in bexI);
apply (rule_tac x = f in bexI);
apply (force simp: OneRequest_def) ;
apply (force simp: RequestQueue_def);
apply blast;

apply (force simp: OneRequest_def) ;

apply ((rule_tac x = A in bexI),(rule_tac x = f in bexI),(force simp: OneRequest_def),(force simp: RequestQueue_def),blast)+;
done;

lemma red_fut_00: "\<lbrakk> c\<rightarrow>\<^sub>\<zeta> c';B\<in>dom c; f \<in> dom (c.\<^sub>RB)\<rbrakk> \<Longrightarrow>  \<exists> B \<in> dom c' . f \<in> dom (c'.\<^sub>RB)"
apply (frule futdom_and_red_inc);
apply (drule_tac x=B in bspec);
apply assumption;
apply (rule_tac x=B in bexI);
apply (force simp: RequestQueue_def)
apply (drule_tac actdom_and_red_inc)
apply force;
done;

(* reduction maintains WF2a+b*)

lemma subject_reduction_WF2[rule_format]: "c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> (wf_2a c\<and>wf_2b c  \<longrightarrow> wf_2a c'\<and>wf_2b c')";
apply (unfold wf_2a_def);
apply (unfold wf_2b_def);
apply (rule impI);
apply (erule conjE);
apply (rule conjI);
apply (rule ballI);
apply (rule allI)+;
apply (rule impI);
apply (frule_tac c=c in futref_and_red_ao);
apply assumption;
apply simp+;
apply (erule disjE, erule bexE);
apply (erule exE);
apply (drule_tac x=B in bspec);
apply assumption;
apply (drule_tac x=E' in spec);
apply (drule_tac x=f in spec);
apply (simp, erule bexE);
apply (rotate_tac -2);
apply (erule red_fut_00,assumption+);

apply (erule bexE)+;
apply (erule disjE);
apply (erule conjE,erule exE);
apply (rotate_tac 2);
apply (drule_tac x=G in bspec);
apply force; 
apply (drule_tac x=f' in bspec);
apply (force simp: RequestQueue_def);
apply (drule_tac x="E'\<circ>(cActive E)" in spec);
apply (drule_tac x=f in spec);
apply simp+; apply (erule bexE);apply (rotate_tac -2); 
apply (erule red_fut_00,assumption+);

apply (erule conjE,(erule exE)+);

apply (rotate_tac 2);
apply (drule_tac x=G in bspec);
apply assumption;
apply (drule_tac x=f' in bspec);
apply assumption;
apply (drule_tac x="E'\<circ>(cUpdR (ActRef D) li E'')" in spec);
apply (drule_tac x=f in spec);
apply (simp add:ectxt2gctxt_and_fill)+;
apply (erule bexE);
apply (rotate_tac -3);apply (erule red_fut_00,assumption+);

apply (rule ballI)+;
apply (rule allI)+;
apply (rule impI);
apply (frule_tac c=c and f=fk and E=E and fk=f in futref_and_red_queue);
apply assumption+;
apply (erule disjE, (erule bexE)+);
apply (erule exE);
apply (rotate_tac 2);
apply (drule_tac x=G in bspec);
apply assumption;
apply (drule_tac x=f' in bspec);
apply assumption;
apply (drule_tac x=E' in spec);
apply (drule_tac x=f in spec);
apply simp+;apply (erule bexE);
apply (rotate_tac -2);
apply (erule red_fut_00,assumption+);

apply (erule disjE)
apply ((erule bexE)+,erule exE);
apply (drule_tac x=G in bspec);
apply assumption;
apply (drule_tac x=E' in spec);
apply (drule_tac x=f in spec);
apply simp+;
apply (erule bexE);apply (rotate_tac -2);
apply (drule_tac f=f in red_fut_00,simp, assumption+);
apply ((erule exE)+,(erule conjE)+);
apply (simp add: fcreated_def);
apply (erule conjE)+;
apply (rotate_tac -3);
apply (rule_tac x="B" in bexI)
apply assumption+;
done;

(* reduction maintains well-formedness *)
theorem wf_preserved:  "\<lbrakk> c \<rightarrow>\<^sub>\<zeta> c'; wf_conf c\<rbrakk> \<Longrightarrow> wf_conf c'";
apply (unfold wf_conf_def);
apply (frule subject_reduction_WF1);
apply simp+;
apply (frule subject_reduction_WF2,simp);
by simp;

lemma fold_rec_forw:
"\<lbrakk> ab_semigroup_mult f; finite A; a: A; fold f g z A \<rbrakk> \<Longrightarrow> f (g a) (fold f g z (A - {a}))"
apply (drule ab_semigroup_mult.fold_rec);
apply assumption+;
apply (erule subst);
by assumption;


lemma and_ab_semigroup_mult: "ab_semigroup_mult op &";
apply (unfold ab_semigroup_mult_def);
apply (unfold semigroup_mult_def ab_semigroup_mult_axioms_def);
apply simp;
apply force;
done;

(* now the particular case of a configuration obtain from initconf, 
  and transitive closure-related theorems*)

lemma noRef_act: " \<not>noRef ((E::general_context)\<up>ActRef A)"
apply (induct_tac E);
apply auto;
apply (subgoal_tac "ab_semigroup_mult op &")
apply (frule_tac a="(FLlabel FmapLabel)" in fold_rec_forw);
prefer 3;
apply (simp add: finite_dom_fmap)+;
apply (force simp: and_ab_semigroup_mult);
done;

lemma noRef_fut: " \<not>noRef ((E::general_context)\<up>FutRef f)"
apply (induct_tac E);
apply auto;
apply (subgoal_tac "ab_semigroup_mult op &")
apply (frule_tac a="(FLlabel FmapLabel)" in fold_rec_forw);
prefer 3;
apply (simp add: finite_dom_fmap)+;
apply (force simp: and_ab_semigroup_mult);
done;

lemma wf_init_conf: "noRef a \<Longrightarrow> wf_conf (init_config a)";
apply (unfold wf_conf_def wf_1a_def wf_1b_def wf_2a_def wf_2b_def init_config_def);
apply simp;
apply (unfold RequestQueue_def OneRequest_def ActiveObject_def);
apply (auto simp: noRef_fut noRef_act);
apply (case_tac E);apply auto;
apply (rotate_tac 1,drule sym, simp add: map_upd_nonempty)
apply (case_tac E);apply auto;
apply (rotate_tac 1,drule sym, simp add: map_upd_nonempty)
done;

lemma wf_conf_rtrancl[rule_format]: "c \<rightarrow>\<^sub>\<zeta>\<^sup>* c' \<Longrightarrow> wf_conf c \<longrightarrow> wf_conf c'";
apply (erule_tac P = "% x y. wf_conf x \<longrightarrow> wf_conf y" in rtranclp.induct);
apply simp;
apply (rule impI)+;
apply (rule wf_preserved);
apply assumption;
apply (erule impE);
apply assumption+;
done;

lemma wf_reachable: "c: reachable_configs \<Longrightarrow> wf_conf c";
apply (unfold reachable_configs_def);
apply simp;
apply (erule exE);
apply (erule conjE);
apply (erule wf_conf_rtrancl);
apply (erule wf_init_conf);
done;

section {* Disjunct request queue domains *}
(* disjunuct futures is true if no two request correspond to the same future 
  (should be the case for a reasonable conf) *)
constdefs 
  disjunct_futures:: "configuration \<Rightarrow> bool"
     "disjunct_futures c == (\<forall>A\<in>dom c. \<forall> B\<in> dom c. (A\<noteq>B) \<longrightarrow> (dom (c.\<^sub>RA)\<inter>dom (c.\<^sub>RB))={})"

(* reduction maintains disjunctness of futures *)
lemma disjunct_futures_SR: "\<lbrakk>c \<rightarrow>\<^sub>\<zeta> c';disjunct_futures c\<rbrakk> \<Longrightarrow> disjunct_futures c'"
  apply (frule_tac P=" disjunct_futures c'" in zeta.cases);
apply (unfold disjunct_futures_def);
(* LOCAL *)
apply (rule ballI)+;
apply (rule impI);
apply (drule_tac x=Aa in bspec);apply (case_tac "Aa=A",force,force);
apply (drule_tac x=B in bspec);apply (case_tac "B=A",force,force);
apply (simp add: RequestQueue_def) 
apply (case_tac "Aa=A",force,force);
(*activity creation *)
apply (rule ballI)+;
apply (rule impI);
apply (case_tac "Aa=G");
apply (simp add: RequestQueue_def) 
apply (drule_tac x=Aa in bspec);apply (case_tac "Aa=A",force,force);
apply (case_tac "B=G");
apply (simp add: RequestQueue_def) 
apply (drule_tac x=B in bspec);apply (case_tac "B=A",force,force);
apply (simp add: RequestQueue_def) 
apply (case_tac "Aa=A",force,force);

(*Request *)
apply (rule ballI)+;
apply (rule impI);
apply (drule_tac x=A in bspec);apply simp+;apply (force);
apply (rotate_tac -1);
apply (drule_tac x=Ba in bspec);apply simp+;apply (force);
apply (case_tac "A=B");
apply (drule_tac x=Ba in bspec);apply simp+;apply (force);
apply (simp add: RequestQueue_def)+;
apply force;
apply (case_tac "Ba=B");
apply (drule_tac x=A in bspec);apply simp+;apply (force);
apply (simp add: RequestQueue_def)+;
apply force;
apply (simp add: RequestQueue_def);
apply force;
(*SELF REQUEST *)
apply (rule ballI)+;
apply (rule impI);
apply (drule_tac x=A in bspec);apply simp+;apply (force);
apply (rotate_tac -1);
apply (drule_tac x=B in bspec);apply simp+;apply (force);
apply (case_tac "A=A'");
apply (drule_tac x=B in bspec);apply simp+;
apply (simp add: RequestQueue_def)+;
apply force;
apply (case_tac "B=A'");
apply (drule_tac x=A in bspec);apply simp+;
apply (simp add: RequestQueue_def)+;
apply force;
apply (simp add: RequestQueue_def);

(*ACctivity update *)
apply (rule ballI)+;
apply (rule impI);
apply (case_tac "A=G");
apply (simp add: RequestQueue_def) 
apply (drule_tac x=A in bspec);apply (simp+,force);
apply (case_tac "Ba=G");
apply (simp add: RequestQueue_def) 
apply (drule_tac x=Ba in bspec);apply (simp+,force);
apply (simp add: RequestQueue_def) 
apply force;

(*REPLY *)
apply (rule ballI)+;
apply (rule impI);
apply (drule_tac x=Aa in bspec);apply (simp+,force);
apply (drule_tac x=Ba in bspec);apply (simp+,force);
apply (simp add: RequestQueue_def) 
apply force;
apply simp+;
done;

lemma initial_config_disjunct: "disjunct_futures (init_config a)"
apply (unfold disjunct_futures_def init_config_def);
by auto;

(* transitive closure extension *)
lemma disjunct_rtrancl[rule_format]: "c \<rightarrow>\<^sub>\<zeta>\<^sup>* c' \<Longrightarrow> disjunct_futures c \<longrightarrow> disjunct_futures c'";
apply (erule_tac P = "% x y. disjunct_futures x \<longrightarrow> disjunct_futures y" in rtranclp.induct);
apply simp;
apply (rule impI)+;
apply (rule disjunct_futures_SR);
apply assumption;
apply (erule impE);
apply assumption+;
done;

(* config reacheble from initconf => disjunct futures *)
theorem reachable_config_disjunct: "c\<in>reachable_configs \<Longrightarrow> disjunct_futures c"
apply (unfold disjunct_futures_def reachable_configs_def);
apply simp+;
apply (erule exE,erule conjE);
apply (subgoal_tac "disjunct_futures (init_config a)");
apply (drule_tac P = "% x y. disjunct_futures x \<longrightarrow> disjunct_futures y" in rtranclp.induct);
apply simp+;
apply (simp add: disjunct_futures_SR);
apply (unfold disjunct_futures_def, simp+, fold disjunct_futures_def);
apply (rule  initial_config_disjunct);
done;
end