 (*  Title:      Config.thy
    ID:         $Id:$
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       Configurations, i.e. parallel activities for a functional
                ASP, based on the sigma-calculus and basic activities 
                defined in ASP.thy
                
                one lemma became an axiom to accelerate the computation time ... proof should work
                
*)

header {* Configurations for ASP *}

theory contexts imports FinASP begin

(* We first define atype for paires of (finite map from labels to dB) 
         x a label of this finite map *)
typedef FmapLabel = "{f::(Label -~> dB) * Label . snd f \<in> dom (fst f)}"
by auto;

(* accessors *)
constdefs
     FLmap:: "FmapLabel \<Rightarrow> (Label -~> dB)"
     "FLmap == fst\<circ>Rep_FmapLabel"
constdefs
     FLlabel:: "FmapLabel \<Rightarrow> Label"
     "FLlabel == snd\<circ>Rep_FmapLabel"

(* the term that corresponds to the label  in the Fmap*)
constdefs
     FLmapsto::"FmapLabel \<Rightarrow> dB"
     "FLmapsto f == the ((FLmap f) (FLlabel f))"

lemma FLdef2: "(FLmap f) (FLlabel f)\<noteq>None"
(* quite difficult compared to what it should be!!! *)
apply (subgoal_tac "(FLlabel f)\<in>dom (FLmap f)");
apply force;
apply (unfold FLmap_def FLlabel_def);
apply (simp add: FmapLabel_def Rep_FmapLabel_inject)+;
apply (insert Rep_FmapLabel);
apply (drule_tac x = f in meta_spec);
apply (insert FmapLabel_def);
apply auto;
done;


lemma FLequals: "\<lbrakk>\<forall>l. FLmap f l = FLmap f' l; FLlabel f = FLlabel f'\<rbrakk>
       \<Longrightarrow> f= f'"
apply (subgoal_tac "FLmap f = FLmap f'");
apply (subgoal_tac "f = Abs_FmapLabel (FLmap f,FLlabel f)"); 
apply (subgoal_tac "f' = Abs_FmapLabel (FLmap f',FLlabel f')"); 
apply force;
apply (unfold FLmap_def);
apply (unfold FLlabel_def);
apply (force simp: Rep_FmapLabel_inverse);
apply (simp (no_asm) add: Rep_FmapLabel_inverse);
apply simp+;
apply (insert expand_fun_eq);
apply (drule_tac x = "fst (Rep_FmapLabel f)" in meta_spec);
apply (drule_tac x = "fst (Rep_FmapLabel f')" in meta_spec);
apply auto;
done;

lemma FLlabelAbs[simp]: "l\<in>dom f \<Longrightarrow> FLlabel( Abs_FmapLabel(f,l)) = l"
apply (auto simp: FLlabel_def); 
apply (insert Abs_FmapLabel_inverse);
apply (simp add: FmapLabel_def);
apply (drule_tac x= "(f,l)" in meta_spec)
apply force;
done;

lemma FLmapAbs[simp]: "l\<in>dom f \<Longrightarrow> FLmap( Abs_FmapLabel(f,l)) = f"
apply (auto simp: FLmap_def); 
apply (insert Abs_FmapLabel_inverse);
apply (simp add: FmapLabel_def);
apply (drule_tac x= "(f,l)" in meta_spec)
apply force;
done;

(* a general context is a term with a hole somewhere *)
datatype general_context =
    cHole
  | cObj FmapLabel type general_context
  | cCallL general_context Label dB 
  | cCallR dB Label general_context 
  | cUpdL general_context Label dB
  | cUpdR dB Label general_context
  | cActive general_context;

(* Fill fills the hole in the general context with a term given in parameter *)
primrec   
       Fill :: "[general_context, dB]  \<Rightarrow>  dB" (infixr "\<up>" 65)
where
    "Fill cHole x = x"
|    "Fill  (cObj f T E ) x =  Obj ((FLmap f)((FLlabel f)\<mapsto>(Fill E x))) T"
|    "Fill  (cCallL E l y) x =  Call (Fill E x) l y"
|    "Fill  (cCallR y l E) x =  Call y l(Fill E x)"
|    "Fill (cUpdL E l (y::dB)) x = Upd (Fill E x) l y"
|    "Fill (cUpdR (y::dB) l E) x = Upd y l (Fill E x)" 
|    "Fill (cActive E) x = Active (Fill E x)"
 
(* econtexts are terms with a hole place in the leftmost position ("classical" evaluation context) *)
(* they are useful because the term placed in the hole o a closed term  has no free variable *)
datatype econtext = (*evaluation context*)
    eHole
  | eCall econtext Label dB
  | eUpdL econtext Label dB
  | eActive econtext 

(* eFill is the equivalent of Fill for econtexts *)
primrec eFill :: "[econtext, dB]  \<Rightarrow>  dB" (infixr "\<up>" 65)
where
    "eFill eHole x = x"
 |   "eFill  (eCall E l b) x  =  Call (eFill E x) l b"
 |   "eFill (eUpdL E l (y::dB)) x = Upd (eFill E x) l y"
 |   "eFill (eActive E) x = Active (eFill E x)"
 

(* cCompose is the composition of contexts: it places the second ctxt in the 
hole of the first ctxtx, this provides a new context *)
primrec cCompose :: "[general_context, general_context]  \<Rightarrow>  general_context" (infixr "\<circ>" 64)
where
    "cCompose cHole F = F"
|    "cCompose (cObj f T E) F =  cObj f T (cCompose E F)"
|    "cCompose (cCallL E l x) F  =  cCallL (cCompose E F) l x"
|    "cCompose (cCallR x l E) F  =  cCallR x l (cCompose E F)"
|    "cCompose (cUpdL E l y) F = cUpdL (cCompose E F) l y"
|    "cCompose (cUpdR y l E) F = cUpdR y l (cCompose E F)" 
|    "cCompose (cActive E) F = cActive (cCompose E F)"

primrec eCompose :: "[econtext, econtext]  \<Rightarrow>  econtext" (infixr "\<circ>" 64)
where
    "eCompose eHole F = F"
|    "eCompose  (eCall E l x) F  =  eCall (eCompose E F) l x"
|    "eCompose (eUpdL E l y) F = eUpdL (eCompose E F) l y"
|    "eCompose (eActive E) F = eActive (eCompose E F)"

(* econtexts are general contexts, the following function does the conversion *)
primrec ectxt2gctxt:: " econtext \<Rightarrow> general_context"
where
    "ectxt2gctxt eHole = cHole"
|    "ectxt2gctxt (eCall E l a) = cCallL (ectxt2gctxt E) l a"
|    "ectxt2gctxt (eUpdL E l a) = cUpdL (ectxt2gctxt E) l a"
|    "ectxt2gctxt (eActive E) = cActive (ectxt2gctxt E)"


section {*csize*}

(* the size of a context -- similar to size for terms; a hole ahas size 0 *) 
primrec csize:: "general_context \<Rightarrow> nat"
where
    "csize cHole = 0"
|    "csize (cObj f T E) =  dBsize (Obj ((FLmap f)(FLlabel f\<mapsto>(Var (Self 0))) ) T)+csize E" 
              (* (Var (Self 0)) is of size 0 => we just want the size of the map with a hole*)
|    "csize (cCallL E l x)  = (csize E) + (dBsize x) + Suc 0"
|    "csize (cCallR x l E)  = (csize E) + (dBsize x) + Suc 0"
|    "csize (cUpdL E l y) = (csize E) + (dBsize y) + Suc 0"
|    "csize (cUpdR y l E) = (dBsize y) + (csize E ) + Suc 0" 
|    "csize (cActive E)  = (csize E) + Suc 0"

primrec esize:: "econtext \<Rightarrow> nat"
where
    "esize eHole = 0"
|    "esize  (eCall E l x)  = (esize E) + (dBsize x)+ Suc 0"
|    "esize (eUpdL E l y) = (esize E) + (dBsize y) + Suc 0"
|    "esize (eActive E)  = (esize E) + Suc 0"

lemma "csize E \<noteq> csize F \<Longrightarrow> E\<noteq>F"
by auto;

(* relates size of terms and field update *)
lemma dBsizeAndReplace: "dBsize (Obj ((FLmap f)(FLlabel f\<mapsto>b) ) T)  =dBsize (Obj ((FLmap f)(FLlabel f\<mapsto>a) ) T) + dBsize b - dBsize a"
apply auto;
apply (subgoal_tac "(FLlabel f)\<in>(dom (FLmap f))")
apply (frule insert_absorb);
apply simp+;
apply (insert comm_monoid_add.times.fold_rec);
apply (drule_tac x = "(dom (FLmap f))" in meta_spec)
apply (drule_tac x = "(FLlabel f)" in meta_spec)
apply (drule_tac x = "(\<lambda>x. dBsize2 (if x = FLlabel f then Some b else FLmap f x))" in meta_spec)
apply (drule_tac x = "Suc 0" in meta_spec);
apply simp;
apply (subgoal_tac "finite (dom (FLmap f))");
apply simp;
apply (insert comm_monoid_add.times.fold_rec);
apply (drule_tac x = "(dom (FLmap f))" in meta_spec)
apply (drule_tac x = "(FLlabel f)" in meta_spec)
apply (drule_tac x = "(\<lambda>x. dBsize2 (if x = FLlabel f then Some a else FLmap f x))" in meta_spec)
apply (drule_tac x = "Suc 0" in meta_spec);
apply simp;
apply (rule comm_monoid_add.times.fold_cong);
apply simp+;
apply (subgoal_tac "dom (FLmap f)\<subseteq> (UNIV :: (Label set))")
apply (insert finite_Label);
apply (rule finite_subset);
apply (insert FLdef2);
apply auto;
done;

(* sized of filled ctxt is the sum of sizes of ctxt + term *)
lemma csize_and_Fill: " (dBsize ((E::general_context)\<up>x) = csize E + dBsize x )"
apply (induct_tac E);
apply simp+;
apply (insert dBsizeAndReplace);
apply (drule_tac x= FmapLabel in meta_spec);
apply (drule_tac x= "general_context \<up> x" in meta_spec);
apply (drule meta_spec);
apply (drule_tac x= "Var (Self 0)" in meta_spec);
apply simp+;
done;
 

section {*Lemmas on contexts*}

lemma ectxt2gctxt_and_fill: "(ectxt2gctxt E)\<up>x = E\<up>x"
apply (induct_tac E);
apply auto;
done;

lemma cCompose_and_Fill: " (\<forall> F x. ((E::general_context)\<circ>F)\<up>x = E\<up>(F\<up>x))"
apply (induct_tac E);
apply auto;
done;

lemma eCompose_and_Fill: " (\<forall> F x. ((E::econtext)\<circ>F)\<up>x = E\<up>(F\<up>x))"
apply (induct_tac E);
apply auto;
done;

lemma size_cHole: "csize (E::general_context)= 0 \<Longrightarrow> E = cHole"
apply (case_tac E);
apply simp;
apply (frule_tac t = "cObj FmapLabel type general_context" and s = E in subst)
apply assumption
apply (subgoal_tac "dBsize (E\<up>(Var (Self 0)))=0");
apply (subgoal_tac "dBsize (E\<up>(Var (Self 0)))>0");
apply simp;
apply (erule ssubst)+;
apply (simp only: Fill.simps);
apply (simp only: SizeOfObjectPos);
apply (simp only: csize_and_Fill);
apply (simp only: dBsize_dBsize2.simps);
apply auto;
done;

lemma fixpoint_context: "x=(E::general_context)\<up>x \<Longrightarrow>E=cHole"
apply (subgoal_tac "csize E = 0");
apply (drule size_cHole);apply assumption;
apply (subgoal_tac "dBsize (E\<up>x) = csize E +dBsize x");
apply simp;
apply (rule csize_and_Fill)
done;

lemma eq_context[rule_format]: " (\<forall> x y .(E::general_context)\<up> x = E\<up> y \<longrightarrow> x = y)"
apply (induct_tac E);
apply auto;
apply (subgoal_tac "(FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> x)) 
                                    (FLlabel FmapLabel) =
        (FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> y)) (FLlabel FmapLabel)")
apply (unfold FLmap_def);
apply (unfold FLlabel_def);
apply (rotate_tac -1);
apply auto;
done;

(* equivalence on t=context is equality, except that the case of cObj because the term corresponding to the label can be anything *)
primrec equiv_contexts:: "general_context \<Rightarrow> general_context \<Rightarrow> bool" (infix "\<approx>" 60)
where
    "equiv_contexts cHole F = (F = cHole)"
|    "equiv_contexts (cObj f T E) F = (\<exists> f' F'.(F = cObj f' T F')\<and> (E\<approx> F')\<and>(FLlabel f = FLlabel f')\<and>
                                      ( \<forall> l .(l\<noteq>FLlabel f \<longrightarrow> ((FLmap f) l = (FLmap f') l))) )"
|    "equiv_contexts  (cCallL E l x) F  =  (\<exists> F' . (F = cCallL F' l x)\<and> (E\<approx> F'))"
|    "equiv_contexts  (cCallR x l E) F  =  (\<exists> F' . (F = cCallR x l F')\<and> (E\<approx> F'))"
|    "equiv_contexts (cUpdL E l y) F = (\<exists> F' . (F = cUpdL F' l y) \<and> (E\<approx>F'))"
|    "equiv_contexts (cUpdR y l E) F = (\<exists> F' . (F = cUpdR y l F')\<and>(E\<approx>F'))" 
|    "equiv_contexts (cActive E) F = (\<exists> F' . (F = cActive F')\<and>(E\<approx>F'))"

lemma equiv_and_equals[rule_format]: " \<forall> E' . (E=E' \<longrightarrow> E\<approx>E')"
apply (induct_tac E);
apply auto;
done;

lemma equiv_and_fill[rule_format]: " \<forall> x E'. (E\<approx>E' \<longrightarrow>E \<up>x = E' \<up> x )"
apply (induct_tac E);
apply auto;
apply (subgoal_tac "\<forall> l .  (FLmap FmapLabel(FLlabel f' \<mapsto> F' \<up> x)) l= (FLmap f'(FLlabel f' \<mapsto> F' \<up> x)) l");
apply (auto simp: expand_fun_eq);
done;

lemma gc_obj_access_filled:
 "(FLmap f(FLlabel f \<mapsto> (E::general_context) \<up> x)) (FLlabel f) =  Some (E \<up> x)";
by auto;

lemma  gcexists_lemma0: "\<lbrakk>\<forall>x y (E'::general_context).
           ((x \<noteq> y \<and> (E::general_context) \<up> x = E' \<up> x \<and> E \<up> y = E' \<up> y) \<longrightarrow> E \<approx> E');
        x \<noteq> y;
        (FLmap f)(FLlabel f \<mapsto> E \<up> x) = (FLmap fa)(FLlabel fa \<mapsto> Ea \<up> x);
        (FLmap f)(FLlabel f \<mapsto> E \<up> y) = (FLmap fa)(FLlabel fa \<mapsto> (Ea::general_context) \<up> y)\<rbrakk>
       \<Longrightarrow> FLlabel f = FLlabel fa"
apply (case_tac "FLlabel f \<noteq> FLlabel fa");
apply (subgoal_tac " (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = 
(FLmap fa (FLlabel fa \<mapsto> Ea \<up> x)) (FLlabel f)");
apply (subgoal_tac "(FLmap f(FLlabel f \<mapsto> E \<up> y))(FLlabel f)
= (FLmap fa (FLlabel fa \<mapsto> Ea \<up> y))(FLlabel f)");
apply (rotate_tac -2);
apply (simp only: fun_upd_def);
apply simp;
apply (drule sym);
apply simp;
apply (insert eq_context);
apply (drule_tac x = E in meta_spec);
apply (rotate_tac -1);
apply (drule_tac x = y in spec);
apply (rotate_tac -1);
apply (drule_tac x=x in spec);
(* this is the fastest end I found *)
apply blast;
apply (rotate_tac -2,simp);
apply fastsimp;
apply blast;
done;

(*axioms general_context_exisits: 
 "  (\<forall> x  y (E'::general_context).x\<noteq>y \<and>(E::general_context)\<up>x = E'\<up>x \<and> E\<up>y = E'\<up>y
                                      \<longrightarrow> E \<approx> E')"; *)
  (*          PROOF (very long below) *)


lemma general_context_exisits: 
 "  (\<forall> x  y (E'::general_context).x\<noteq>y \<and>(E::general_context)\<up>x = E'\<up>x \<and> E\<up>y = E'\<up>y
                                       \<longrightarrow> E \<approx> E')";
apply (induct_tac E);
apply simp;
apply clarify;
apply (drule fixpoint_context);apply simp+;
apply (rule allI)+;
apply (case_tac E');
apply (auto simp: fixpoint_context);

apply (subgoal_tac "x=(cObj FmapLabel type general_context)\<up>x ");
apply (drule fixpoint_context);
apply simp;
apply (simp add: fixpoint_context);
apply (subgoal_tac "(FLlabel FmapLabel) = (FLlabel FmapLabela)");
apply (drule_tac x=x in spec);
apply (drule_tac x = y in spec );
apply (drule_tac x = general_contexta in spec );
apply simp+;
apply (insert gc_obj_access_filled);
apply (drule_tac x = FmapLabel in meta_spec);
apply (drule_tac x = general_context in meta_spec);
apply (drule_tac x = x in meta_spec);
apply (insert gc_obj_access_filled);
apply (drule_tac x = FmapLabel in meta_spec);
apply (drule_tac x = general_context in meta_spec);
apply (drule_tac x = y in meta_spec);
apply (insert gc_obj_access_filled);
apply (drule_tac x = FmapLabel in meta_spec);
apply (drule_tac x = general_contexta in meta_spec);
apply (drule_tac x = x in meta_spec);
apply (insert gc_obj_access_filled);
apply (drule_tac x = FmapLabel in meta_spec);
apply (drule_tac x = general_contexta in meta_spec);
apply (drule_tac x = y in meta_spec);
apply simp;

apply (rule gcexists_lemma0);
apply simp;
apply simp;
apply simp;
apply simp;

apply (rule gcexists_lemma0);
apply simp;
apply simp;
apply simp;
apply simp;


apply (case_tac "FLlabel FmapLabel \<noteq> FLlabel FmapLabela");
apply (subgoal_tac " (FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> x)) (FLlabel FmapLabel) = 
(FLmap FmapLabela (FLlabel FmapLabela \<mapsto> general_contexta \<up> x)) (FLlabel FmapLabel)");
apply (subgoal_tac "(FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> y))(FLlabel FmapLabel)
= (FLmap FmapLabela (FLlabel FmapLabela \<mapsto> general_contexta \<up> y))(FLlabel FmapLabel)");
apply (rotate_tac -2);
apply (simp only: fun_upd_def);
apply simp;
apply (drule sym);
apply simp;
apply (insert eq_context)
apply (drule_tac x = general_context in meta_spec);
apply (drule_tac x = y in spec);
apply (drule_tac x = y in spec);
apply blast;

apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (thin_tac "\<And>f E x. (FLmap f(FLlabel f \<mapsto> E \<up> x)) (FLlabel f) = Some ((E::general_context) \<up> x)")+;
defer;
apply (subgoal_tac " (FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> x)) l = 
(FLmap FmapLabela (FLlabel FmapLabela \<mapsto> general_contexta \<up> x)) l");
apply (rotate_tac -1)

apply (simp only: fun_upd_def);
apply fastsimp;
apply simp;
(* 6 *)
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'");

apply (rotate_tac 1);
apply (drule_tac f="\<lambda>x. (x l) " in arg_cong)
apply fastsimp;
(* 5 *)

apply (case_tac E'); 
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y");
apply (simp,rotate_tac 1,  drule sym);
apply (cut_tac x=x and  E="(cCallL general_context Label  dB)" in fixpoint_context);
apply simp;apply simp;
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y");
apply simp;
apply (drule_tac x=x in spec);
apply (drule_tac x=y in spec);
apply (drule_tac x=general_contexta in spec);
apply simp;

apply (drule_tac x=general_context in meta_spec);
apply (drule_tac x=x in meta_spec);
apply simp;

apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;
(* 4 *)
apply (case_tac E');
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y");
apply (cut_tac x=x and  E="(cCallR dB Label  general_context)" in fixpoint_context);
apply simp;apply simp;
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;
apply (drule_tac x=general_context in meta_spec);
apply (drule_tac x=x in meta_spec);
apply simp;

apply (drule_tac x=x in spec);
apply (drule_tac x=y in spec);
apply (drule_tac x=general_contexta in spec);
apply simp;

apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;

(* 3 *)

apply (case_tac E');
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y");
apply (cut_tac x=x and  E="(cUpdL general_context Label dB)" in fixpoint_context);
apply simp;apply simp;
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;

apply (drule_tac x=x in spec);
apply (drule_tac x=y in spec);
apply (drule_tac x=general_contexta in spec);
apply simp;

apply (drule_tac x=general_context in meta_spec);
apply (drule_tac x=x in meta_spec);
apply simp;

apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;

(* 2 *)

apply (case_tac E');
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y");
apply (cut_tac x=x and  E="(cUpdR dB Label  general_context)" in fixpoint_context);
apply simp;apply simp;
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;
apply (drule_tac x=general_context in meta_spec);
apply (drule_tac x=x in meta_spec);
apply simp;

apply (drule_tac x=x in spec);
apply (drule_tac x=y in spec);
apply (drule_tac x=general_contexta in spec);
apply simp;

apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;

(* 1 *)

apply (case_tac E');
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y");
apply (cut_tac x=x and  E="(cActive general_context)" in fixpoint_context);
apply simp;apply simp;
apply (thin_tac "\<forall>x y E'. x \<noteq> y \<and> general_context \<up> x = E' \<up> x \<and> general_context \<up> y = E' \<up> y \<longrightarrow> general_context \<approx> E'", thin_tac "\<And>E x y. (E::general_context) \<up> x = E \<up> y \<Longrightarrow> x = y",fastsimp)+;

apply (drule_tac x=x in spec);
apply (drule_tac x=y in spec);
apply (drule_tac x=general_contexta in spec);
apply simp;
done;


(* useful consequence of the equality of 2 filled contexts *)
lemma contexts_and_dB[rule_format]: " (\<forall> x y (E'::general_context). (E::general_context)\<up>x = E'\<up>y \<longrightarrow>
                ((\<exists> (E1::general_context). x =E1\<up>y)\<or>(\<exists> (E1::general_context). E1\<up>x =y)\<or>
                           (\<forall>z. \<exists> (E2::general_context). E\<up> z = E2\<up> y)))"
apply (induct_tac E);
apply simp+;
apply (rule allI)+;
apply (rule disjI1);
apply (rule_tac x = E' in exI);
apply simp+;
apply (rule allI)+;
apply (rule impI)+;
apply (erule exE);
apply (case_tac E');
apply simp+;
apply (rotate_tac -2);
apply (drule sym);
apply simp+;
apply (rule disjI2);
apply (rule disjI1);
apply (rule_tac x = "cObj FmapLabel type general_context" in exI);
apply simp+;
apply (case_tac "FLlabel FmapLabel = FLlabel FmapLabela");
apply simp+;
apply (erule conjE);
apply (drule_tac f = "\<lambda>x. (x (FLlabel FmapLabela))" in arg_cong);
apply (simp only: fun_upd_def);
apply (rotate_tac -1);
apply (simp);
apply (erule_tac x = x in allE);
apply (erule_tac x = y in allE);
apply simp+;
apply (drule mp);
apply force;
apply (erule disjE);
apply simp+;
apply (erule disjE);
apply simp+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (erule_tac x = z in allE);
apply (erule exE)+;
apply (rule_tac x = "cObj FmapLabel typea (E2)" in exI);
apply simp;
apply (simp only: fun_upd_def);
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (rule_tac x = "cObj (Abs_FmapLabel(FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> z),FLlabel FmapLabela)) typea general_contexta" in exI);
apply simp;
apply (insert Abs_FmapLabel_inverse);
apply (drule_tac x = "(fst (Rep_FmapLabel FmapLabela)(snd (Rep_FmapLabel FmapLabela) \<mapsto>
                   general_context \<up> z), snd (Rep_FmapLabel FmapLabelaa))" in meta_spec)
apply (drule impI, drule mp);
apply (simp add: FmapLabel_def);
apply (rule disjI2);
apply (erule conjE);
apply (drule_tac f = "\<lambda>x. (x (FLlabel FmapLabela))" in arg_cong);
apply (simp add: fun_upd_def)+;
apply (rotate_tac 2, drule not_sym);
apply (simp add: FLmap_def FLlabel_def)+;
apply force;

apply (simp add: FLmap_def FLlabel_def)+;
apply (subgoal_tac "\<forall> l . (fst (Rep_FmapLabel FmapLabel)(snd (Rep_FmapLabel FmapLabel) \<mapsto>
          general_context \<up> z)) l = 
(fst (Rep_FmapLabel FmapLabel)(snd (Rep_FmapLabel FmapLabel) 
          \<mapsto> general_context \<up> z, snd (Rep_FmapLabel FmapLabela) \<mapsto> general_contexta \<up> y) ) l");
apply (simp add: expand_fun_eq);
apply (rule allI);
apply (case_tac "l = snd (Rep_FmapLabel FmapLabela)"); 
apply simp+;
apply (rotate_tac -3, drule not_sym);
apply (erule conjE);
apply (drule_tac f = "\<lambda>x. (x (snd (Rep_FmapLabel FmapLabela)))" in arg_cong);
apply simp;
apply force;apply simp+;

(* callL *)
apply (rule allI)+;
apply (rule impI);
apply (erule exE);
apply (case_tac E');
apply simp_all;
apply (rule disjI2);
apply (rule disjI1);
apply (rule_tac x = "cCallL general_context Label dB" in exI);
apply simp+;
apply (erule conjE);
apply (drule_tac x = x in spec);
apply (drule_tac x = y in spec);
apply (drule mp);
apply (rule_tac x = "general_contexta" in exI);
apply simp+;
apply (erule disjE, simp)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (erule_tac x = z in allE);
apply (erule exE)+;
apply simp+;
apply (rule_tac x = "cCallL (E2) Labela dBa" in exI);
apply simp+;
apply (erule conjE)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (rule_tac x = "cCallR (general_context \<up> z) Labela general_contexta" in exI);
apply simp+;

(* callR *)
apply (rule allI)+;
apply (rule impI);
apply (erule exE);
apply (case_tac E');
apply simp_all;
apply (rule disjI2);
apply (rule disjI1);
apply (rule_tac x = "cCallR dB Label general_context" in exI);
apply simp+;
apply (erule conjE)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (rule_tac x = "cCallL  general_contexta Labela (general_context \<up> z)" in exI);
apply simp+;
apply (erule conjE);
apply (drule_tac x = x in spec);
apply (drule_tac x = y in spec);
apply (drule mp);
apply (rule_tac x = "general_contexta" in exI);
apply simp+;
apply (erule disjE, simp)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (erule_tac x = z in allE);
apply (erule exE)+;
apply simp+;
apply (rule_tac x = "cCallR dBa Labela E2" in exI);
apply simp+;

(* UpdL *)
apply (rule allI)+;
apply (rule impI);
apply (erule exE);
apply (case_tac E');
apply simp_all;
apply (rule disjI2);
apply (rule disjI1);
apply (rule_tac x = "cUpdL general_context Label dB" in exI);
apply simp+;
apply (erule conjE);
apply (drule_tac x = x in spec);
apply (drule_tac x = y in spec);
apply (drule mp);
apply (rule_tac x = "general_contexta" in exI);
apply simp+;
apply (erule disjE, simp)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (erule_tac x = z in allE);
apply (erule exE)+;
apply simp+;
apply (rule_tac x = "cUpdL (E2) Labela dBa" in exI);
apply simp+;
apply (erule conjE)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (rule_tac x = "cUpdR (general_context \<up> z) Labela general_contexta" in exI);
apply simp+;

(* UpdR *)
apply (rule allI)+;
apply (rule impI);
apply (erule exE);
apply (case_tac E');
apply simp+;
apply (rule disjI2);
apply (rule disjI1);
apply (rule_tac x = "cUpdR dB Label general_context" in exI);
apply simp+;
apply (erule conjE)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (rule_tac x = "cUpdL  general_contexta Labela (general_context \<up> z)" in exI);
apply simp+;
apply (erule conjE);
apply (drule_tac x = x in spec);
apply (drule_tac x = y in spec);
apply (drule mp);
apply (rule_tac x = "general_contexta" in exI);
apply simp+;
apply (erule disjE, simp)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (erule_tac x = z in allE);
apply (erule exE)+;
apply simp+;
apply (rule_tac x = "cUpdR dBa Labela E2" in exI);
apply simp+;

(* cActive *)
apply (rule allI)+;
apply (rule impI);
apply (erule exE);
apply (case_tac E');
apply simp+;
apply (rule disjI2);
apply (rule disjI1);
apply (rule_tac x = "cActive general_context" in exI);
apply simp+;
apply (drule_tac x = x in spec);
apply (drule_tac x = y in spec);
apply (drule mp);
apply (rule_tac x = "general_contexta" in exI);
apply simp+;
apply (erule disjE, simp)+;
apply (rule disjI2);
apply (rule disjI2);
apply (rule allI);
apply (erule_tac x = z in allE);
apply (erule exE)+;
apply simp+;
apply (rule_tac x = "cActive E2" in exI);
apply simp+;
done; 

(* FORMATTED : lemma contexts_and_dB:
        ?E \<up> ?x = ?E' \<up> ?y \<Longrightarrow>
        (\<exists>E1. ?x = E1 \<up> ?y) \<or> (\<exists>E1. E1 \<up> ?x = ?y) \<or> (\<forall>z. \<exists>E2. ?E \<up> z = E2 \<up> ?y)   *)

(* if filling gives a Call *)
lemma context_call_1[rule_format]: " \<forall> x li d b. (((E::general_context)\<up>x) = Call d li b\<longrightarrow>
                                       ((E = cHole \<or> (\<exists> (E'::general_context). d = E' \<up> x))\<or> (\<exists> (E'::general_context). b = E' \<up> x)))";
apply (case_tac E);
apply auto;
done;
(*formated: context_call_1: ?E \<up> ?x = Call ?d ?li \<Longrightarrow> ?E = cHole \<or> (\<exists>E'. ?d = E' \<up> ?x)*)

(* if filling results in Upd *)
lemma context_update[rule_format]: " \<forall> x li a b. (((E::general_context)\<up>x) = Upd a li b\<longrightarrow>
                                       (E = cHole \<or> (\<exists> (E'::general_context). a = E' \<up> x))\<or> (\<exists> (E'::general_context). b = E' \<up> x))";
apply (case_tac E);
apply auto;
done;

section {* ActRef specific lemmas*}

lemma lift_context_Act: "(\<forall> (E::general_context)  n G. 
                               (lift x n = (E \<up> (ActRef G)) \<longrightarrow> (\<exists> (E'::general_context) .  x = E' \<up> (ActRef G)))) \<and>
  (\<forall> (E::general_context)  n G.(lift_option n xo) = Some (E \<up> (ActRef G)) \<longrightarrow> (\<exists> (E'::general_context). xo = Some (E' \<up> (ActRef G))))"
apply (induct_tac x and xo);
apply auto;
apply (case_tac E, simp+);
apply (case_tac E, simp_all);
defer;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = n in spec);
apply (drule_tac x = G in spec);
apply simp;
apply (erule exE);
apply (rule_tac x = "cCallL E' Labela dB2" in exI);
apply simp+
apply (erule conjE)+;
apply (rotate_tac 1);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = n in spec);
apply (drule_tac x = G in spec);
apply simp;
apply (erule exE);
apply (rule_tac x = "cCallR dB1 Labela E'" in exI);
apply simp+;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = n in spec);
apply (drule_tac x = G in spec);
apply simp;
apply (erule exE);
apply (rule_tac x = "cUpdL E' Labela dB2" in exI);
apply simp+;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec)+;
apply (drule_tac x = "(Suc n)" in spec)+;
apply (drule_tac x = G in spec)+;
apply simp;
apply clarify;
apply (rule_tac x = "cUpdR dB1 Labela E'" in exI);
apply simp+;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp;
apply clarify;
apply (rule_tac x = "cActive  E'" in exI);
apply simp+;
apply (erule conjE);
apply (frule_tac f = "\<lambda>f. (f (FLlabel FmapLabel))" in arg_cong);
apply (drule_tac x = "(FLlabel FmapLabel)" in meta_spec);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "Suc n" in spec);
apply (drule_tac x = G in spec);
apply simp;
apply clarify;
apply (rule_tac x = "cObj (Abs_FmapLabel (fun,(FLlabel FmapLabel)
)) typea E'" in exI)
apply simp+;
apply (simp add: fun_upd_def FLmap_def FLlabel_def );
apply (subgoal_tac "Rep_FmapLabel (Abs_FmapLabel(fun,snd (Rep_FmapLabel FmapLabel)) ) 
                    = (fun,    snd (Rep_FmapLabel FmapLabel))");
apply (rotate_tac -1);
apply (simp add: fun_upd_def FLmap_def FLlabel_def )+;
apply (rule_tac s = "(fun, snd (Rep_FmapLabel FmapLabel))" in ssubst);
apply assumption;
apply (subgoal_tac " \<forall> l . fun l= (\<lambda>x. if x = snd (Rep_FmapLabel FmapLabel) then Some (E' \<up> ActRef G)
                     else fst (fun, snd (Rep_FmapLabel FmapLabel)) x) l");
apply (simp add: expand_fun_eq);
apply force;
apply (rule Abs_FmapLabel_inverse);
apply (auto simp: FmapLabel_def);
done;

lemma lift_context_Act_1: 
"lift x n = (E::general_context) \<up> (ActRef G) \<Longrightarrow> \<exists> (E'::general_context) .  x = E' \<up> (ActRef G)"
apply (auto simp: lift_context_Act);
done

lemma subst_context_Act: 
"(\<forall> E b c n G.( a[b,c/n]=(E::general_context)\<up>(ActRef G) 
  \<longrightarrow> (\<exists> E'::general_context . a=E' \<up>(ActRef G)) 
      \<or>(\<exists> E'::general_context .b = E' \<up>(ActRef G))
      \<or>(\<exists> E'::general_context .c = E' \<up>(ActRef G)))) \<and>
 (\<forall> (E::general_context)  b c n G. ( subst_option n b c ao=Some (E\<up>(ActRef G)) 
  \<longrightarrow> (\<exists> E'::general_context . ao= Some (E' \<up>(ActRef G))) 
      \<or>(\<exists> E'::general_context .b = E' \<up>(ActRef G))
      \<or>(\<exists> E'::general_context .c = E' \<up>(ActRef G)) ) )
"
apply (induct_tac a and ao);
apply (rule allI)+; apply (rule impI);

(* Var *)
apply (case_tac Variable);

(* Self *)
apply (case_tac "n=nat");
apply (simp add: split_if);
apply blast;
apply (case_tac "n<nat");
apply simp;
apply (subgoal_tac "E=cHole");
apply force;
apply (case_tac E);apply simp+;
apply force;

(* Param *)
apply (case_tac "n=nat");
apply (simp add: split_if);
apply blast;
apply (case_tac "n<nat");
apply simp;
apply (subgoal_tac "E=cHole");
apply force;
apply (case_tac E);apply simp+;
apply force;

(* Obj *)
apply (rule allI)+; apply (rule impI);
apply simp+;
apply (case_tac E);
apply simp+;

apply (drule_tac x = "(FLlabel FmapLabel)" in meta_spec);
apply (drule_tac x = "general_context" in spec);
apply (drule_tac x = "(lift b (0\<Colon>nat))" in spec);
apply (drule_tac x = "(lift c (0\<Colon>nat))" in spec);
apply (drule_tac x = "Suc n" in spec);
apply (drule_tac x = G in spec);
apply (erule conjE);
apply (frule_tac f = "\<lambda>f. (f (FLlabel FmapLabel))" in arg_cong);
apply simp+;
apply (erule disjE);
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = "cObj (Abs_FmapLabel (fun, FLlabel FmapLabel)) typea E'" in exI)
apply simp+;
apply (simp add: FLlabel_def fun_upd_def);
apply (subgoal_tac "(Rep_FmapLabel (Abs_FmapLabel (fun, snd (Rep_FmapLabel FmapLabel)))) = (fun, snd (Rep_FmapLabel FmapLabel))");
apply (subgoal_tac " \<forall>l . fun l = (\<lambda>x. if x = snd (Rep_FmapLabel (Abs_FmapLabel (fun, snd (Rep_FmapLabel FmapLabel))))
                     then Some (E' \<up> ActRef G) else FLmap (Abs_FmapLabel (fun, snd (Rep_FmapLabel FmapLabel))) x) l");
apply (simp add: expand_fun_eq);
apply  (simp add: FLmap_def)+;
apply (rule Abs_FmapLabel_inverse);
apply (simp add: FmapLabel_def)+;
apply force;
apply (rule disjI2);
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (erule lift_context_Act_1);
apply (rule disjI2);
apply (erule exE);
apply (erule lift_context_Act_1);
apply simp+;

(* Call *)
apply (rule allI)+; apply (rule impI);
apply (case_tac E);
apply simp+;
apply (erule conjE);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp+;
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (rule_tac x = "cCallL E' Labela dB2" in exI);apply simp;
apply (rule disjI2);
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply force;
apply (rule disjI2);
apply (erule exE);
apply force;
apply simp+;
apply (rotate_tac 1);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp+;
apply (erule disjE);
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = "cCallR dB1 Labela E'" in exI);apply simp+;

(* Upd *)
apply (rule allI)+; apply (rule impI);
apply (case_tac E);
apply simp+;
apply (erule conjE);
apply (drule_tac x = general_context in spec)+;
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec)+;
apply (drule_tac x = G in spec)+;
apply simp+;
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (rule_tac x = "cUpdL E' Labela dB2" in exI);apply simp+;
apply (drule_tac x = general_context in spec)+;
apply (drule_tac x = "lift b (0\<Colon>nat)" in spec);
apply (drule_tac x = "lift b (0\<Colon>nat)" in spec);
apply (drule_tac x = "lift c (0\<Colon>nat)" in spec)+;
apply (drule_tac x = "Suc n" in spec)+;
apply (drule_tac x = G in spec)+;
apply simp+;
apply (erule disjE);
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = "cUpdR dB1 Labela E'" in exI);apply simp+;
apply (rule disjI2);
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (simp add: lift_context_Act_1);
apply (rule disjI2);
apply (erule exE);
apply (simp add: lift_context_Act_1);
apply simp+;

(* Active *)
apply (rule allI)+; apply (rule impI);
apply (case_tac E);
apply simp+;
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp+;
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (rule_tac x = "cActive E'" in exI);apply simp+;
apply auto;
done;

 lemma subst_context_Act_1: 
" a[b,c/n]=(E::general_context)\<up>(ActRef G)
    \<Longrightarrow> ((\<exists> E'::general_context . a=E' \<up>(ActRef G))  
       \<or> (\<exists> E'::general_context .b = E' \<up>(ActRef G))
       \<or> (\<exists> E'::general_context .c = E' \<up>(ActRef G)))"
by (auto simp: subst_context_Act);

lemma context_inc_Act: " (E::general_context) \<up> a = (E'::general_context) \<up> ActRef G 
  \<Longrightarrow> ((\<exists> E1::general_context. a = E1 \<up> ActRef G) \<or> (\<forall>b.\<exists> E2::general_context. E\<up>b=E2\<up>ActRef G))"
apply (drule contexts_and_dB);
apply (erule disjE);
apply simp;
apply (erule disjE);
apply (erule exE);
apply (case_tac E1);
apply (rule disjI1);
apply (rule_tac x=cHole in exI);
apply auto;
done;




section {* FutRef specific lemmas*}

lemma lift_context_Fut: "(\<forall> (E::general_context)  n G. 
                               (lift x n = (E \<up> (FutRef G)) \<longrightarrow> (\<exists> (E'::general_context) .  x = E' \<up> (FutRef G)))) \<and>
  (\<forall> (E::general_context)  n G.(lift_option n xo) = Some (E \<up> (FutRef G)) \<longrightarrow> (\<exists> (E'::general_context). xo = Some (E' \<up> (FutRef G))))"
apply (induct_tac x and xo);
apply auto;
apply (case_tac E, simp+);
apply (case_tac E, simp_all);
defer;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = n in spec);
apply (drule_tac x = G in spec);
apply simp;
apply (erule exE);
apply (rule_tac x = "cCallL E' Labela dB2" in exI);
apply simp+
apply (erule conjE)+;
apply (rotate_tac 1);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = n in spec);
apply (drule_tac x = G in spec);
apply simp;
apply (erule exE);
apply (rule_tac x = "cCallR dB1 Labela E'" in exI);
apply simp+;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = n in spec);
apply (drule_tac x = G in spec);
apply simp;
apply (erule exE);
apply (rule_tac x = "cUpdL E' Labela dB2" in exI);
apply simp+;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec)+;
apply (drule_tac x = "(Suc n)" in spec)+;
apply (drule_tac x = G in spec)+;
apply simp;
apply clarify;
apply (rule_tac x = "cUpdR dB1 Labela E'" in exI);
apply simp+;
apply (case_tac E, simp_all);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp;
apply clarify;
apply (rule_tac x = "cActive  E'" in exI);
apply simp+;
apply (erule conjE);
apply (frule_tac f = "\<lambda>f. (f (FLlabel FmapLabel))" in arg_cong);
apply (drule_tac x = "(FLlabel FmapLabel)" in meta_spec);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "Suc n" in spec);
apply (drule_tac x = G in spec);
apply simp;
apply clarify;
apply (rule_tac x = "cObj (Abs_FmapLabel (fun,(FLlabel FmapLabel)
)) typea E'" in exI)
apply simp+;
apply (simp add: fun_upd_def FLmap_def FLlabel_def );
apply (subgoal_tac "Rep_FmapLabel (Abs_FmapLabel(fun,snd (Rep_FmapLabel FmapLabel)) ) 
                    = (fun,    snd (Rep_FmapLabel FmapLabel))");
apply (rotate_tac -1);
apply (simp add: fun_upd_def FLmap_def FLlabel_def )+;
apply (rule_tac s = "(fun, snd (Rep_FmapLabel FmapLabel))" in ssubst);
apply assumption;
apply (subgoal_tac " \<forall> l . fun l= (\<lambda>x. if x = snd (Rep_FmapLabel FmapLabel) then Some (E' \<up> FutRef G)
                     else fst (fun, snd (Rep_FmapLabel FmapLabel)) x) l");
apply (simp add: expand_fun_eq);
apply force;
apply (rule Abs_FmapLabel_inverse);
apply (auto simp: FmapLabel_def);
done;

lemma lift_context_Fut_1: 
"lift x n = (E::general_context) \<up> (FutRef G) \<Longrightarrow> \<exists> (E'::general_context) .  x = E' \<up> (FutRef G)"
apply (auto simp: lift_context_Fut);
done

lemma subst_context_Fut: 
"(\<forall> E b c n G.( a[b,c/n]=(E::general_context)\<up>(FutRef G) 
  \<longrightarrow> (\<exists> E'::general_context . a=E' \<up>(FutRef G)) 
      \<or>(\<exists> E'::general_context .b = E' \<up>(FutRef G))
      \<or>(\<exists> E'::general_context .c = E' \<up>(FutRef G)))) \<and>
 (\<forall> (E::general_context)  b c n G. ( subst_option n b c ao=Some (E\<up>(FutRef G)) 
  \<longrightarrow> (\<exists> E'::general_context . ao= Some (E' \<up>(FutRef G))) 
      \<or>(\<exists> E'::general_context .b = E' \<up>(FutRef G))
      \<or>(\<exists> E'::general_context .c = E' \<up>(FutRef G)) ) )
"
apply (induct_tac a and ao);
apply (rule allI)+; apply (rule impI);

(* Var *)
apply (case_tac Variable);

(* Self *)
apply (case_tac "n=nat");
apply (simp add: split_if);
apply blast;
apply (case_tac "n<nat");
apply simp;
apply (subgoal_tac "E=cHole");
apply force;
apply (case_tac E);apply simp+;
apply force;

(* Param *)
apply (case_tac "n=nat");
apply (simp add: split_if);
apply blast;
apply (case_tac "n<nat");
apply simp;
apply (subgoal_tac "E=cHole");
apply force;
apply (case_tac E);apply simp+;
apply force;

(* Obj *)
apply (rule allI)+; apply (rule impI);
apply simp+;
apply (case_tac E);
apply simp+;

apply (drule_tac x = "(FLlabel FmapLabel)" in meta_spec);
apply (drule_tac x = "general_context" in spec);
apply (drule_tac x = "(lift b (0\<Colon>nat))" in spec);
apply (drule_tac x = "(lift c (0\<Colon>nat))" in spec);
apply (drule_tac x = "Suc n" in spec);
apply (drule_tac x = G in spec);
apply (erule conjE);
apply (frule_tac f = "\<lambda>f. (f (FLlabel FmapLabel))" in arg_cong);
apply simp+;
apply (erule disjE);
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = "cObj (Abs_FmapLabel (fun, FLlabel FmapLabel)) typea E'" in exI)
apply simp+;
apply (simp add: FLlabel_def fun_upd_def);
apply (subgoal_tac "(Rep_FmapLabel (Abs_FmapLabel (fun, snd (Rep_FmapLabel FmapLabel)))) = (fun, snd (Rep_FmapLabel FmapLabel))");
apply (subgoal_tac " \<forall>l . fun l = (\<lambda>x. if x = snd (Rep_FmapLabel (Abs_FmapLabel (fun, snd (Rep_FmapLabel FmapLabel))))
                     then Some (E' \<up> FutRef G) else FLmap (Abs_FmapLabel (fun, snd (Rep_FmapLabel FmapLabel))) x) l");
apply (simp add: expand_fun_eq);
apply  (simp add: FLmap_def)+;
apply (rule Abs_FmapLabel_inverse);
apply (simp add: FmapLabel_def)+;
apply force;
apply (rule disjI2);
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (erule lift_context_Fut_1);
apply (rule disjI2);
apply (erule exE);
apply (erule lift_context_Fut_1);
apply simp+;

(* Call *)
apply (rule allI)+; apply (rule impI);
apply (case_tac E);
apply simp+;
apply (erule conjE);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp+;
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (rule_tac x = "cCallL E' Labela dB2" in exI);apply simp;
apply (rule disjI2);
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply force;
apply (rule disjI2);
apply (erule exE);
apply force;
apply simp+;
apply (rotate_tac 1);
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp+;
apply (erule disjE);
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = "cCallR dB1 Labela E'" in exI);apply simp+;

(* Upd *)
apply (rule allI)+; apply (rule impI);
apply (case_tac E);
apply simp+;
apply (erule conjE);
apply (drule_tac x = general_context in spec)+;
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec)+;
apply (drule_tac x = G in spec)+;
apply simp+;
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (rule_tac x = "cUpdL E' Labela dB2" in exI);apply simp+;
apply (drule_tac x = general_context in spec)+;
apply (drule_tac x = "lift b (0\<Colon>nat)" in spec);
apply (drule_tac x = "lift b (0\<Colon>nat)" in spec);
apply (drule_tac x = "lift c (0\<Colon>nat)" in spec)+;
apply (drule_tac x = "Suc n" in spec)+;
apply (drule_tac x = G in spec)+;
apply simp+;
apply (erule disjE);
apply (erule exE);
apply (rule disjI1);
apply (rule_tac x = "cUpdR dB1 Labela E'" in exI);apply simp+;
apply (rule disjI2);
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (simp add: lift_context_Fut_1);
apply (rule disjI2);
apply (erule exE);
apply (simp add: lift_context_Fut_1);
apply simp+;

(* Active *)
apply (rule allI)+; apply (rule impI);
apply (case_tac E);
apply simp+;
apply (drule_tac x = general_context in spec);
apply (drule_tac x = "b" in spec);
apply (drule_tac x = "c" in spec);
apply (drule_tac x = "n" in spec);
apply (drule_tac x = G in spec);
apply simp+;
apply (erule disjE);
apply (rule disjI1);
apply (erule exE);
apply (rule_tac x = "cActive E'" in exI);
apply auto;
done;

lemma subst_context_Fut_1: 
" a[b,c/n]=(E::general_context)\<up>(FutRef G)
    \<Longrightarrow> ((\<exists> E'::general_context . a=E' \<up>(FutRef G))  
       \<or> (\<exists> E'::general_context .b = E' \<up>(FutRef G))
       \<or> (\<exists> E'::general_context .c = E' \<up>(FutRef G)))"
by (auto simp: subst_context_Fut);

lemma context_inc_Fut: " (E::general_context) \<up> a = (E'::general_context) \<up> FutRef G 
  \<Longrightarrow> ((\<exists> E1::general_context. a = E1 \<up> FutRef G) \<or> (\<forall>b.\<exists> E2::general_context. E\<up>b=E2\<up>FutRef G))"
apply (drule contexts_and_dB);
apply (erule disjE);
apply simp;
apply (erule disjE);
apply (erule exE);
apply (case_tac E1);
apply (rule disjI1);
apply (rule_tac x=cHole in exI);
apply auto;
done;


section {* econtext lemmas *}

lemma eq_econtext: " (E::econtext)\<up> x = E\<up> y \<Longrightarrow> x = y"
apply (subgoal_tac "(ectxt2gctxt E)\<up>x = (ectxt2gctxt E)\<up>y")
apply (erule eq_context)
apply (simp add: ectxt2gctxt_and_fill);
done;

end
